"""Structure-wise describers. These describers include structural information."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from maml.base import BaseDescriber, describer_type

if TYPE_CHECKING:
    from pymatgen.core import Molecule, Structure

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


__all__ = [
    "CoulombEigenSpectrum",
    "CoulombMatrix",
    "DistinctSiteProperty",
    "RandomizedCoulombMatrix",
    "SortedCoulombMatrix",
]


@describer_type("structure")
class DistinctSiteProperty(BaseDescriber):
    r"""
    Constructs a describers based on properties of distinct sites in a
    structure. For now, this assumes that there is only one type of species in
    a particular Wyckoff site.

    Reference:
    @article{ye2018deep,
            title={Deep neural networks for accurate predictions of crystal stability},
            author={Ye, Weike and Chen, Chi and Wang, Zhenbin and
                Chu, Iek-Heng and Ong, Shyue Ping},
            journal={Nature communications},
            volume={9},
            number={1},
            pages={1--6},
            year={2018},
            publisher={Nature Publishing Group}}
    """

    # todo: generalize to multiple sites with the same Wyckoff.
    supported_properties = [
        "mendeleev_no",
        "electrical_resistivity",
        "velocity_of_sound",
        "reflectivity",
        "refractive_index",
        "poissons_ratio",
        "molar_volume",
        "thermal_conductivity",
        "boiling_point",
        "melting_point",
        "critical_temperature",
        "superconduction_temperature",
        "liquid_range",
        "bulk_modulus",
        "youngs_modulus",
        "brinell_hardness",
        "rigidity_modulus",
        "mineral_hardness",
        "vickers_hardness",
        "density_of_solid",
        "atomic_radius_calculated",
        "van_der_waals_radius",
        "coefficient_of_linear_thermal_expansion",
        "ground_state_term_symbol",
        "valence",
        "Z",
        "X",
        "atomic_mass",
        "block",
        "row",
        "group",
        "atomic_radius",
        "average_ionic_radius",
        "average_cationic_radius",
        "average_anionic_radius",
        "metallic_radius",
        "ionic_radii",
        "oxi_state",
        "max_oxidation_state",
        "min_oxidation_state",
        "is_transition_metal",
        "is_alkali",
        "is_alkaline",
        "is_chalcogen",
        "is_halogen",
        "is_lanthanoid",
        "is_metal",
        "is_metalloid",
        "is_noble_gas",
        "is_post_transition_metal",
        "is_quadrupolar",
        "is_rare_earth_metal",
        "is_actinoid",
    ]

    def __init__(
        self,
        properties: list[str],
        symprec: float = 0.1,
        wyckoffs: list[str] | None = None,
        feature_batch: str = "pandas_concat",
        **kwargs,
    ):
        """

        Args:
            properties (list): Sequence of specie properties. E.g.,
                ["atomic_radius"]. Look at pymatgen.core.periodic_table.Element and
                pymatgen.core.periodic_table.Species for support properties (there
                are a lot!)
            symprec (float): Symmetry precision for spacegroup determination.
            wyckoffs (list of wyckoff symbols):. E.g., ["48a", "24c"], if not provided,
                will get from input structure when doing transform
            feature_batch (str): way to batch a list of features into one
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        self.properties = properties
        self.symprec = symprec
        self.wyckoffs = wyckoffs
        super().__init__(feature_batch=feature_batch, **kwargs)

    def transform_one(self, structure: Structure) -> pd.DataFrame:
        """

        Args:
            structure (pymatgen Structure): pymatgen structure for descriptor computation.

        Returns:
            pd.DataFrame that contains the distinct position labeled features

        """
        a = SpacegroupAnalyzer(structure, self.symprec)
        symm = a.get_symmetrized_structure()
        data = []
        names = []
        if self.wyckoffs is None:
            self.wyckoffs = symm.wyckoff_symbols
        for w in self.wyckoffs:
            site = symm.equivalent_sites[symm.wyckoff_symbols.index(w)][0]
            for p in self.properties:
                data.append(getattr(site.specie, p))
                names.append(f"{w}-{p}")
        return pd.DataFrame([data], columns=names)


@describer_type("structure")
class CoulombMatrix(BaseDescriber):
    r"""
    Coulomb Matrix to describe structure.

    Reference:
    @article{rupp2012fast,
            title={Fast and accurate modeling of molecular
                atomization energies with machine learning},
            author={Rupp, Matthias and Tkatchenko, Alexandre and M{\"u}ller,
                Klaus-Robert and Von Lilienfeld, O Anatole},
            journal={Physical review letters}, volume={108},
            number={5}, pages={058301},
            year={2012}, publisher={APS}}
    """

    def __init__(self, random_seed: int | None = None, max_atoms: int | None = None, is_ravel: bool = True, **kwargs):
        """
        Args:
            random_seed (int): random seed
            max_atoms (int): maximum number of atoms
            is_ravel (bool): whether to ravel the matrix to 1D
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        self.max_atoms = max_atoms
        self.random_seed = random_seed
        self.is_ravel = is_ravel
        if "feature_batch" not in kwargs:
            kwargs["feature_batch"] = "pandas_concat"
        super().__init__(**kwargs)

    @staticmethod
    def _get_columb_mat(s: Molecule | Structure) -> np.ndarray:
        """
        Args:
            s (Molecule/Structure): input Molecule or Structure. Structure
                is not advised since the feature will depend on the supercell size.

        Returns:
            Coulomb matrix of the structure

        """
        dis = s.distance_matrix
        np.fill_diagonal(dis, np.inf)  # avoid dividing by zero
        zs = np.array([i.specie.Z for i in s])
        z_matrix = zs[:, None] * zs[None, :]
        z_diag = 0.5 * zs**2.4
        c = z_matrix / dis
        np.fill_diagonal(c, z_diag)
        return c

    def get_coulomb_mat(self, s: Molecule | Structure) -> np.ndarray:
        """
        Args:
            s (Molecule/Structure): input Molecule or Structure. Structure
                is not advised since the feature will depend on the supercell size
        Returns:
            Coulomb matrix of the structure.

        """
        c = self._get_columb_mat(s)
        num_sites = c.shape[0]
        if self.max_atoms is not None and self.max_atoms > num_sites:
            padding = self.max_atoms - num_sites
            return np.pad(c, (0, padding), mode="constant", constant_values=0)
        return c

    def transform_one(self, s: Molecule | Structure) -> pd.DataFrame:
        """
        Args:
            s (Molecule/Structure): pymatgen Molecule or Structure, Structure is not
                advised since the features will depend on supercell size.

        Returns:
            pandas.DataFrame.
            The column is index of the structure, which is 0 for single input
            df[0] returns the serials of coulomb_mat raval
        """
        c = self.get_coulomb_mat(s)
        if self.is_ravel:
            c = c.ravel()
        return pd.DataFrame(c)


@describer_type("structure")
class RandomizedCoulombMatrix(CoulombMatrix):
    r"""
    Randomized CoulombMatrix.

    Reference:
    @article{montavon2013machine,
            title={Machine learning of molecular electronic properties
                in chemical compound space},
            author={Montavon, Gr{\'e}goire and Rupp, Matthias and Gobre,
                Vivekanand and Vazquez-Mayagoitia, Alvaro and Hansen, Katja
                and Tkatchenko, Alexandre and M{\"u}ller, Klaus-Robert and
                Von Lilienfeld, O Anatole},
            journal={New Journal of Physics},
            volume={15}, number={9},pages={095003},
            year={2013},publisher={IOP Publishing}}
    """

    def __init__(self, random_seed: int | None = None, is_ravel: bool = True, **kwargs):
        """
        Args:
            random_seed (int): random seed
            is_ravel (bool): Passthrough to is_ravel for CoulombMatrix. Defaults to True.
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        super().__init__(random_seed=random_seed, is_ravel=is_ravel, **kwargs)

    def get_randomized_coulomb_mat(self, s: Molecule | Structure) -> pd.DataFrame:
        """
        Returns the randomized matrix
        (i) take an arbitrary valid Coulomb matrix C
        (ii) compute the norm of each row of this Coulomb matrix: row_norms
        (iii) draw a zero-mean unit-variance noise vector ε of the same
            size as row_norms.
        (iv)  permute the rows and columns of C with the same permutation
            that sorts row_norms + ε.

        Args:
            s (Molecule/Structure): pymatgen Molecule or Structure, Structure is not
                advised since the features will depend on supercell size

        Returns:
            pd.DataFrame randomized Coulomb matrix
        """
        c = self.get_coulomb_mat(s)
        row_norms = np.linalg.norm(c, axis=1)
        rng = np.random.RandomState(self.random_seed)  # pylint: disable=E1101
        e = rng.normal(size=row_norms.size)
        p = np.argsort(row_norms + e)
        c = c[p][:, p]
        if self.is_ravel:
            c = c.ravel()
        return pd.DataFrame(c)

    def transform_one(self, s: Molecule | Structure) -> pd.DataFrame:
        """
        Transform one structure to descriptors
        Args:
            s (Molecule/Structure): pymatgen Molecule or Structure, Structure is not
                advised since the features will depend on supercell size.

        Returns: pandas dataframe descriptors

        """
        return self.get_randomized_coulomb_mat(s)


@describer_type("structure")
class SortedCoulombMatrix(CoulombMatrix):
    r"""
    Sorted CoulombMatrix.

    Reference:
    @inproceedings{montavon2012learning,
                title={Learning invariant representations
                    of molecules for atomization energy prediction},
                author={Montavon, Gr{\'e}goire and Hansen, Katja
                    and Fazli, Siamac and Rupp, Matthias and Biegler,
                    Franziska and Ziehe, Andreas and Tkatchenko, Alexandre
                    and Lilienfeld, Anatole V and M{\"u}ller, Klaus-Robert},
                booktitle={Advances in neural information processing systems},
                pages={440--448}, year={2012}}
    """

    def __init__(self, random_seed: int | None = None, is_ravel: bool = True, **kwargs):
        """
        Args:
            random_seed (int): random seed
            is_ravel (bool): Passthrough to is_ravel for CoulombMatrix. Defaults to True.
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        super().__init__(random_seed=random_seed, is_ravel=is_ravel, **kwargs)

    def get_sorted_coulomb_mat(self, s: Molecule | Structure) -> pd.DataFrame:
        """
        Returns the matrix sorted by the row norm.

        Args:
            s (Molecule/Structure): pymatgen Molecule or Structure, Structure is not
                advised since the features will depend on supercell size

        Returns:
            pd.DataFrame, sorted Coulomb matrix
        """
        c = self.get_coulomb_mat(s)
        c = c[np.argsort(np.linalg.norm(c, axis=1))]
        if self.is_ravel:
            c = c.ravel()
        return pd.DataFrame(c)

    def transform_one(self, s: Molecule | Structure) -> pd.DataFrame:
        """
        Transform one structure into descriptor
        Args:
            s (Molecule/Structure): pymatgen Molecule or Structure, Structure is not
                advised since the features will depend on supercell size.

        Returns: pd.DataFrame descriptors

        """
        return self.get_sorted_coulomb_mat(s)


@describer_type("structure")
class CoulombEigenSpectrum(BaseDescriber):
    r"""
    Get the Coulomb Eigen Spectrum describers.

    Reference:
    @article{rupp2012fast,
            title={Fast and accurate modeling of molecular
                atomization energies with machine learning},
            author={Rupp, Matthias and Tkatchenko, Alexandre and M{\"u}ller,
                Klaus-Robert and Von Lilienfeld, O Anatole},
            journal={Physical review letters}, volume={108},
            number={5}, pages={058301},
            year={2012}, publisher={APS}}
    """

    def __init__(self, max_atoms: int | None = None, **kwargs):
        """
        This method calculates the Coulomb matrix of a molecule and
        then sort the eigen values of the Coulomb matrix as the vector
        features for the molecule.
        When multiple molecules are converted at the same time, the
        describers will stack the results. If the number of atoms is
        not the same for molecules, zeros will padded to the features.

        Args:
            max_atoms (int): maximum number of atoms
            **kwargs: Passthrough to parent class __init__.
        """
        feature_batch = "stack_padded" if max_atoms is None else "stack_first_dim"
        self.max_atoms = max_atoms
        super().__init__(feature_batch=feature_batch, **kwargs)

    def transform_one(self, mol: Molecule) -> np.ndarray:
        """
        Args:
            mol (Molecule): pymatgen molecule.

        Returns: np.ndarray the eigen value vectors of Coulob matrix

        """
        c_mat = CoulombMatrix._get_columb_mat(mol)
        eig_vals = np.linalg.eigvals(c_mat)
        eig_vals = np.abs(eig_vals)

        f = np.sort(eig_vals)[::-1]
        if self.max_atoms is not None:
            if self.max_atoms < len(f):
                raise RuntimeError("max_atoms is smaller than the size of current molecule")
            f = np.pad(f, (0, self.max_atoms - len(f)))
        return f



"""
Sure Independence Screening.

https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf

"""

from __future__ import annotations

import logging
from itertools import combinations

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.metrics import get_scorer

from maml.apps.symbolic._selectors import BaseSelector, DantzigSelector

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _get_coeff(x, y):
    coeff, _, _, _ = np.linalg.lstsq(x, y, rcond=-1)
    return coeff


def _eval(x, y, coeff, metric):
    metric_func = get_scorer(metric)
    lr = LinearRegression(fit_intercept=False)
    lr.coef_ = coeff  # type: ignore
    lr.intercept_ = 0
    return metric_func(lr, x, y)


def _best_combination(x, y, find_sel, find_sel_new, metric: str = "neg_mean_absolute_error"):
    if len(find_sel_new) == 1:
        comb_best = np.append(find_sel, find_sel_new)
        coeff_best = _get_coeff(x[:, comb_best], y)
        score_best = _eval(x[:, comb_best], y, coeff_best, metric)
        return comb_best, coeff_best, score_best
    combs = combinations(np.append(find_sel, find_sel_new), len(find_sel) + 1)
    coeff_best = _get_coeff(x[:, find_sel], y)
    score_best = _eval(x[:, find_sel], y, coeff_best, metric)
    comb_best = find_sel
    for ind_comb in combs:
        d = x[:, ind_comb]
        coeff = _get_coeff(d, y)
        score = _eval(d, y, coeff, metric)
        if score > score_best:
            score_best = score
            comb_best = ind_comb  # type: ignore
            coeff_best = coeff
    return comb_best, coeff_best, score_best


class SIS:
    """
    Sure independence screening method.
    The method consists of two steps:
        1. Screen
        2. Select.

    """

    def __init__(self, gamma=0.1, selector: BaseSelector | None = None, verbose: bool = True):
        """
        Sure independence screening.

        Args:
            gamma (float): ratio between selected features and original feature sizes
            selector (BaseSelector): selector after the screening
            verbose (bool): whether to output information along the way

        """
        self.gamma = gamma
        self.selector = selector
        self.verbose = verbose

    def run(self, x, y, select_options=None):
        """
        Run the SIS with selector
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            select_options (dict): options in the optimizations provided
                to scipy.optimize.minimize. If the selector is using cvxpy
                optimization package, this option is fed into cp.Problem.solve.

        Returns: selected feature indices

        """
        screened_indices = self.screen(x, y)
        if self.verbose:
            logger.info(f"After the screening step, {len(screened_indices)}/{x.shape[1]} features remains")
        x_screen = x[:, screened_indices]

        final_selected = self.select(x_screen, y, select_options)

        if self.verbose:
            logger.info(f"After the selection step, {len(final_selected)}/{x.shape[1]} features remains")
        return screened_indices[final_selected]

    def screen(self, x, y):
        """
        Simple screening method by comparing the correlation between features
        and the target.

        Args:
            x (np.ndarray): input array
            y (np.ndarray): target array

        Returns: top indices

        """
        n = x.shape[1]
        omega = x.T.dot(y)
        sorted_omega = np.argsort(omega)[::-1]
        d = int(n * self.gamma)
        return sorted_omega[:d]

    def select(self, x, y, options=None):
        """
        Select features using selectors
        Args:
            x (np.ndarray): input array
            y (np.ndarray): target array
            options (dict): options for the optimization.
        """
        return self.selector.select(x, y, options)

    def compute_residual(self, x, y):
        """
        Compute residual
        Args:
            x (np.ndarray): input array
            y (np.ndarray): target array.

        Returns: residual vector

        """
        return self.selector.compute_residual(x, y)

    def set_selector(self, selector: BaseSelector):
        """
        Set new selector
        Args:
            selector (BaseSelector): a feature selector.
        """
        self.selector = selector

    def set_gamma(self, gamma):
        """
        Set gamma.

        Args:
            gamma (float): new gamma value

        """
        self.gamma = gamma

    def update_gamma(self, ratio: float = 0.5):
        """
        Update the sis object so that sis.select
        return at least one feature.

        Args:
            ratio (float): ratio to update the parameters

        """
        self.set_gamma(self.gamma * (1 + ratio))


class ISIS:
    """Iterative SIS."""

    def __init__(self, sis: SIS | None = None, l0_regulate: bool = True):
        """

        Args:
            sis(SIS): sis object
            l0_regulate(bool): Whether to regulate features in each iteration, default True.
        """
        self.sis = sis if sis is not None else SIS(gamma=0.1, selector=DantzigSelector(0.1))
        self.selector = sis.selector  # type: ignore
        self.l0_regulate = l0_regulate
        self.coeff = []  # type: ignore
        self.find_sel = []  # type: ignore

    def run(
        self,
        x: np.ndarray,
        y: np.ndarray,
        max_p: int = 10,
        metric: str = "neg_mean_absolute_error",
        options: dict | None = None,
        step: float = 0.5,
    ):
        """
        Run the ISIS
        Args:
            x (np.ndarray): input array
            y (np.ndarray): target array
            max_p(int): Number of feature desired
            metric (str): scorer function, used with
                sklearn.metrics.get_scorer
            options:
            step(float): step to update gamma with.

        Returns:
            find_sel(np.array): np.array of index of selected features
            coeff(np.array): np.array of coeff of selected features

        """
        assert max_p <= x.shape[1]
        findex = np.array(np.arange(0, x.shape[1]))
        find_sel = self.sis.select(x, y, options)
        self.coeff = _get_coeff(x[:, find_sel], y)

        if len(find_sel) >= max_p:
            self.coeff = _get_coeff(x[:, find_sel[:max_p]], y)
            return find_sel[:max_p]
        new_findex = np.array(list(set(findex) - set(find_sel)))
        new_y = self.sis.compute_residual(x, y)
        new_x = x[:, new_findex]
        while len(find_sel) < max_p:
            find_sel_new: list[int] = []
            try:
                find_sel_new = self.sis.run(new_x, new_y, options)
            except ValueError:
                while len(find_sel_new) == 0:
                    self.sis.update_gamma(step)
                    find_sel_new = self.sis.run(new_x, new_y)
            if self.l0_regulate:
                find_sel, _, _ = _best_combination(x, y, find_sel, new_findex[find_sel_new], metric)
            else:
                find_sel = np.append(find_sel, new_findex[find_sel_new])
            new_findex = np.array(list(set(findex) - set(find_sel)))
            new_y = self.sis.compute_residual(new_x, new_y)
            new_x = x[:, new_findex]
        self.coeff = _get_coeff(x[:, find_sel], y)
        self.find_sel = find_sel
        return find_sel

    def evaluate(self, x: np.ndarray, y: np.ndarray, metric: str = "neg_mean_absolute_error") -> float:
        """
        Evaluate the linear models using x, and y test data.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            metric (str): scorer function, used with
                sklearn.metrics.get_scorer
        Returns: float.
        """
        return _eval(x[:, self.find_sel], y, self.coeff, metric)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from maml.base import BaseDescriber, describer_type

if TYPE_CHECKING:
    from pymatgen.core import Molecule, Structure

DEFAULT_MODEL = Path(__file__).parent / "data/m3gnet_models/matbench_mp_e_form/0/m3gnet/"


@describer_type("structure")
class M3GNetStructure(BaseDescriber):
    """Use M3GNet pre-trained models as featurizer to get Structural features."""

    def __init__(
        self,
        model_path: str | None = None,
        **kwargs,
    ):
        """

        Args:
            model_path (str): m3gnet models path. If no path is provided,
                the models will be M3GNet formation energy model on figshare:
                https://figshare.com/articles/software/m3gnet_property_model_weights/20099465
                Please refer to the M3GNet paper:
                https://doi.org/10.1038/s43588-022-00349-3.
            **kwargs: Pass through to BaseDescriber.
        """
        from m3gnet.models import M3GNet

        if model_path:
            self.describer_model = M3GNet.from_dir(model_path)
            self.model_path = model_path
        else:
            self.describer_model = M3GNet.from_dir(DEFAULT_MODEL)
            self.model_path = str(DEFAULT_MODEL)
        super().__init__(**kwargs)

    def transform_one(self, structure: Structure | Molecule):
        """
        Transform structure/molecule objects into structural features.

        Args:
            structure (Structure/Molecule): target object structure or molecule
        Returns: M3GNet readout layer output as structural features.

        """
        from m3gnet.graph import Index, tf_compute_distance_angle
        from m3gnet.layers import polynomial

        graph = self.describer_model.graph_converter.convert(structure).as_list()
        graph = tf_compute_distance_angle(graph)
        three_basis = self.describer_model.basis_expansion(graph)
        three_cutoff = polynomial(graph[Index.BONDS], self.describer_model.threebody_cutoff)
        g = self.describer_model.featurizer(graph)
        g = self.describer_model.feature_adjust(g)
        for i in range(self.describer_model.n_blocks):
            g = self.describer_model.three_interactions[i](g, three_basis, three_cutoff)
            g = self.describer_model.graph_layers[i](g)
        layer_before_readout = self.describer_model.layers[-2].layers[0]
        return np.array(layer_before_readout(g))[0]


@describer_type("site")
class M3GNetSite(BaseDescriber):
    """Use M3GNet pre-trained models as featurizer to get atomic features."""

    def __init__(
        self,
        model_path: str | None = None,
        output_layers: list | None = None,
        return_type: list | dict = pd.DataFrame,
        **kwargs,
    ):
        """

        Args:
            model_path (str): m3gnet models path. If no path is provided,
                the models will be M3GNet formation energy model on figshare:
                https://figshare.com/articles/software/m3gnet_property_model_weights/20099465
                Please refer to the M3GNet paper:
                https://doi.org/10.1038/s43588-022-00349-3.
            output_layers: List of names for the layer of GNN as output. Choose from "embedding", "gc_1", "gc_2",...,
                "gc_n", where n is the total number of graph convolutional layers. By default, the node features in
                "gc_1" layer are returned.
            return_type: The data type of the returned the atom features. By default, atom features in different
                output_layers are concatenated to one vector per atom, and a dataframe of vectors are returned.
            **kwargs: Pass through to BaseDescriber. E.g., feature_batch="pandas_concat" is very useful (see test).
        """
        from m3gnet.models import M3GNet

        if model_path:
            self.describer_model = M3GNet.from_dir(model_path)
            self.model_path = model_path
        else:
            self.describer_model = M3GNet.from_dir(DEFAULT_MODEL)
            self.model_path = str(DEFAULT_MODEL)
        allowed_output_layers = ["embedding"] + [f"gc_{i + 1}" for i in range(self.describer_model.n_blocks)]
        if output_layers is None:
            output_layers = ["gc_1"]
        elif not isinstance(output_layers, list) or set(output_layers).difference(allowed_output_layers):
            raise ValueError(f"Invalid output_layers, it must be a sublist of {allowed_output_layers}.")
        self.output_layers = output_layers
        self.return_type = return_type
        super().__init__(**kwargs)

    def transform_one(self, structure: Structure | Molecule):
        """
        Transform structure/molecule objects into atom features
        Args:
            structure (Structure/Molecule): target object structure or molecule
        Returns: M3GNet node features as atom features.

        """
        from m3gnet.graph import Index, tf_compute_distance_angle
        from m3gnet.layers import polynomial

        graph = self.describer_model.graph_converter.convert(structure).as_list()
        graph = tf_compute_distance_angle(graph)
        three_basis = self.describer_model.basis_expansion(graph)
        three_cutoff = polynomial(graph[Index.BONDS], self.describer_model.threebody_cutoff)
        g = self.describer_model.featurizer(graph)
        atom_fea = {"embedding": g[Index.ATOMS]}
        g = self.describer_model.feature_adjust(g)
        for i in range(self.describer_model.n_blocks):
            g = self.describer_model.three_interactions[i](g, three_basis, three_cutoff)
            g = self.describer_model.graph_layers[i](g)
            atom_fea[f"gc_{i + 1}"] = g[Index.ATOMS]
        atom_fea_dict = {k: v for k, v in atom_fea.items() if k in self.output_layers}
        if isinstance(self.return_type, dict):
            return atom_fea_dict
        return pd.DataFrame(np.concatenate(list(atom_fea_dict.values()), axis=1))



"""Module implements the perturbation class for atomic and lattice relaxation."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

import numpy as np
from monty.serialization import loadfn
from pymatgen.core.structure import Lattice, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.groups import SpaceGroup, in_array_list

if TYPE_CHECKING:
    from pymatgen.core.operations import SymmOp
    from pymatgen.core.sites import PeriodicSite, Site

module_dir = os.path.dirname(__file__)

wyckoff_nums = loadfn(os.path.join(module_dir, "symmetry_rules", "wyckoff_nums.json"))
wyckoff_nums = {int(k): v for k, v in wyckoff_nums.items()}

wyckoff_dims = loadfn(os.path.join(module_dir, "symmetry_rules", "wyckoff_dims.json"))
wyckoff_dims = {int(k): v for k, v in wyckoff_dims.items()}

standard_modes = loadfn(os.path.join(module_dir, "symmetry_rules", "standard_modes.json"))
standard_modes = {int(k): v for k, v in standard_modes.items()}

perturbation_modes = loadfn(os.path.join(module_dir, "symmetry_rules", "perturbation_modes.json"))
perturbation_modes = {int(k): v for k, v in perturbation_modes.items()}

small_addup = np.array([1e-4] * 3)


def perturbation_mapping(x, fixed_indices):
    """
    Perturbation mapping.

    Args:
        x: Perturbation mapping for unfixed lattice.
        fixed_indices: indices to fix.

    Returns: Perturbation mapping.

    """
    return np.array(
        [
            0 if i in fixed_indices else x[np.argwhere(np.arange(3)[~np.isin(range(3), fixed_indices)] == i)[0][0]]
            for i in range(3)
        ]
    )


class WyckoffPerturbation:
    """
    Perturbation class for determining the standard wyckoff position
    and generating corresponding equivalent fractional coordinates.
    """

    def __init__(
        self,
        int_symbol: int,
        wyckoff_symbol: str,
        symmetry_ops: list[SymmOp] | None = None,
        use_symmetry: bool = True,
    ):
        """
        Args:
            int_symbol (int): International number of space group.
            wyckoff_symbol (str): Wyckoff symbol.
            symmetry_ops (list): Full set of symmetry operations as matrices.
                Use specific symmetry operations if initialized.
            use_symmetry (bool): Whether to use constraint of symmetry to reduce
                parameters space.
        """
        self._site = None
        self._fit_site = False
        self.int_symbol = int_symbol
        self.wyckoff_symbol = wyckoff_symbol
        self.use_symmetry = use_symmetry
        if self.use_symmetry:
            self.standard_mode = eval(standard_modes[int_symbol][wyckoff_symbol])
            self.dim = wyckoff_dims[int_symbol][wyckoff_symbol]
            self.multiplicity = dict(zip(*wyckoff_nums[int_symbol], strict=False))[wyckoff_symbol]
            self.perturbation_mode = eval(perturbation_modes[int_symbol][wyckoff_symbol])
            self.symmetry_ops = symmetry_ops if symmetry_ops else SpaceGroup.from_int_number(int_symbol).symmetry_ops
        else:
            self.standard_mode = eval("lambda p: True")
            self.dim = 3
            self.multiplicity = 1
            self.perturbation_mode = eval("lambda x: x")
            self.symmetry_ops = SpaceGroup.from_int_number(1).symmetry_ops

    def get_orbit(self, p: list | np.ndarray, tol: float = 1e-3) -> list[np.ndarray]:
        """
        Returns the orbit for a point.

        Args:
            p (list/numpy.array): Fractional coordinated point.
            tol (float): Tolerance for determining if sites are the same.
        """
        orbit: list[np.ndarray] = []
        for symm_op in self.symmetry_ops:
            pp = symm_op.operate(p)
            pp[(pp + np.ones(3) * tol) % 1.0 < tol] = 0.0
            pp = np.mod(np.round(pp, decimals=10), 1)
            if not in_array_list(orbit, pp, tol=tol):
                orbit.append(pp)

        return orbit

    def sanity_check(self, site: Site | PeriodicSite, wyc_tol: float = 0.3 * 1e-3) -> None:
        """
        Check whether the perturbation mode exists.

        Args:
            site (PeriodicSite): PeriodicSite in Structure.
            wyc_tol (float): Tolerance for wyckoff symbol determined coordinates.
        """
        p = site.frac_coords
        orbits = self.get_orbit(p, wyc_tol)

        if len(orbits) != self.multiplicity:
            return

        for pp in orbits:
            if self.standard_mode(pp):
                self._site = site  # type: ignore
                self._fit_site = True
                break

    def standardize(self, p: list | np.ndarray, tol: float = 1e-3) -> list[float]:
        """
        Get the standardized position of p.

        Args:
            p (list/numpy.array): Fractional coordinated point.
            tol (float): Tolerance for determining if sites are the same.
        """
        pp: list[float] = []
        orbits = self.get_orbit(p, tol)
        for pp in orbits:  # type: ignore
            if self.standard_mode(pp):
                break
        return pp

    @property
    def site(self):
        """Returns the site."""
        return self._site

    @property
    def fit_site(self):
        """Returns whether the site fits any standard wyckoff position."""
        return self._fit_site

    def __repr__(self):
        if self._site is not None:
            a, b, c = self._site.frac_coords
            return (
                f"{self.__class__.__name__}(spg_int_number={self.int_symbol}, wyckoff_symbol={self.wyckoff_symbol})"
                f" {self._site.species_string} [{a:.4f}, {b:.4f}, {c:.4f}]"
            )
        return f"{self.__class__.__name__}(spg_int_number={self.int_symbol}, wyckoff_symbol={self.wyckoff_symbol})"


def crystal_system(int_number: int) -> str:
    """
    Method for crystal system determination.

    Args:
        int_number (int): International number of space group.
    """
    if int_number <= 2:
        return "triclinic"
    if int_number <= 15:
        return "monoclinic"
    if int_number <= 74:
        return "orthorhombic"
    if int_number <= 142:
        return "tetragonal"
    if int_number <= 167 and int_number not in [
        143,
        144,
        145,
        147,
        149,
        150,
        151,
        152,
        153,
        154,
        156,
        157,
        158,
        159,
        162,
        163,
        164,
        165,
    ]:
        return "rhombohedral"
    if int_number <= 194:
        return "hexagonal"
    return "cubic"


class LatticePerturbation:
    """Perturbation class for determining the standard lattice."""

    def __init__(self, spg_int_symbol: int, use_symmetry: bool = True):
        """
        Args:
            spg_int_symbol (int): International number of space group.
            use_symmetry (bool): Whether to use constraint of symmetry to reduce
                parameters space.
        """
        self._lattice = None
        self._fit_lattice = False
        self.spg_int_symbol = spg_int_symbol
        self.use_symmetry = use_symmetry
        self.crys_system = crystal_system(spg_int_symbol)

    def sanity_check(self, lattice: Lattice, abc_tol: float = 1e-3, angle_tol: float = 3e-1) -> None:
        """
        Check whether the perturbation mode exists.

        Args:
            lattice (Lattice): Lattice in Structure.
            abc_tol (float): Tolerance for lattice lengths determined by crystal system.
            angle_tol (float): Tolerance for lattice angles determined by crystal system.
        """
        abc = list(lattice.abc)
        angles = list(lattice.angles)

        def check(param, ref, tolerance):
            return all(abs(i - j) < tolerance for i, j in zip(param, ref, strict=False) if j is not None)

        if not self.use_symmetry:
            self.dims = (3, 3)
            self.perturbation_mode = lambda x: x
            self._lattice = lattice  # type: ignore
            self._abc = abc
            self._fit_lattice = True
            return

        if self.crys_system == "cubic":
            a = abc[0]
            if not (check(abc, [a, a, a], abc_tol) and check(angles, [90, 90, 90], angle_tol)):
                self._fit_lattice = False
                return
            self.dims = (1, 0)
            self.perturbation_mode = lambda x: np.concatenate((np.repeat(x, repeats=3), np.zeros(3)))
            self._lattice = lattice  # type: ignore
            self._abc = [abc[0]]  # type: ignore
            self._fit_lattice = True
            return
        if self.crys_system == "hexagonal":
            if not (
                np.any([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)])
                and check(np.sort(angles), [90, 90, 120], angle_tol)
            ):
                self._fit_lattice = False
                return
            self.dims = (2, 0)
            indices = [int(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]
            self.perturbation_mode = lambda x: np.concatenate((x[indices], np.zeros(3)))
            self._lattice = lattice  # type: ignore
            self._abc = [
                abc[indices.index(0)],  # type: ignore
                abc[indices.index(1)],
            ]  # type: ignore
            self._fit_lattice = True
            return
        if self.crys_system == "rhombohedral":
            a = abc[0]
            alpha = angles[0]
            if check(abc, [a, a, a], abc_tol) and check(angles, [alpha, alpha, alpha], angle_tol):
                self.dims = (1, 1)
                self.perturbation_mode = lambda x: np.concatenate(
                    (np.repeat(x[0], repeats=3), np.repeat(x[1], repeats=3))
                )
                self._lattice = lattice  # type: ignore
                self._abc = [a]  # type: ignore
                self._fit_lattice = True
                return
            if np.any([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]) and check(
                np.sort(angles), [90, 90, 120], angle_tol
            ):
                self.dims = (2, 0)
                indices = [int(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]
                self.perturbation_mode = lambda x: np.concatenate((x[indices], np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = [
                    abc[indices.index(0)],  # type: ignore
                    abc[indices.index(1)],
                ]  # type: ignore
                self._fit_lattice = True
                return
            self._fit_lattice = False
            return
        if self.crys_system == "tetragonal":
            if not check(angles, [90, 90, 90], angle_tol):
                self._fit_lattice = False
                return
            if np.any([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]):
                self.dims = (2, 0)
                indices = [int(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]
                self.perturbation_mode = lambda x: np.concatenate((x[indices], np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = [
                    abc[indices.index(0)],  # type: ignore
                    abc[indices.index(1)],
                ]  # type: ignore
                self._fit_lattice = True
                return
            if np.all([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 3) for i in np.arange(3)]):
                self.dims = (1, 0)
                self.perturbation_mode = lambda x: np.concatenate((np.repeat(x, repeats=3), np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = [abc[0]]  # type: ignore
                self._fit_lattice = True
                return
            self._fit_lattice = False
            return
        if self.crys_system == "orthorhombic":
            if not check(angles, [90, 90, 90], angle_tol):
                self._fit_lattice = False
                return
            self.dims = (3, 0)
            self.perturbation_mode = lambda x: np.concatenate((x, np.zeros(3)))
            self._lattice = lattice  # type: ignore
            self._abc = list(abc)  # type: ignore
            self._fit_lattice = True
            return
        if self.crys_system == "monoclinic":
            if sum(abs(angles[i] - 90) < 1e-5 for i in np.arange(3)) == 2:
                self.dims = (3, 1)
                indices = [int(abs(angles[i] - 90) < angle_tol) for i in np.arange(3)]
                self.perturbation_mode = lambda x: np.concatenate(
                    (
                        np.pad(x, pad_width=(0, 5 - len(x)), mode="constant")[:3],
                        np.pad(x, pad_width=(0, 5 - len(x)), mode="constant")[3:][indices],
                    )
                )
                self._lattice = lattice  # type: ignore
                self._abc = list(abc)  # type: ignore
                self._fit_lattice = True
                return
            if check(angles, [90, 90, 90], angle_tol):
                self.dims = (3, 0)
                self.perturbation_mode = lambda x: np.concatenate((x, np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = list(abc)  # type: ignore
                self._fit_lattice = True
                return
            self._fit_lattice = False
            return
        self.dims = (3, 3)
        self.perturbation_mode = lambda x: x
        self._lattice = lattice  # type: ignore
        self._abc = list(abc)  # type: ignore
        self._fit_lattice = True
        return

    @property
    def fit_lattice(self) -> bool:
        """Returns whether the lattice fits any crystal system."""
        return self._fit_lattice

    @property
    def lattice(self) -> Lattice:
        """Returns the lattice."""
        return self._lattice  # type: ignore

    @property
    def abc(self) -> list[float]:
        """Returns the lattice lengths."""
        return self._abc  # type: ignore

    def __repr__(self):
        if self._lattice is not None:
            return (
                f"{self.__class__.__name__}(spg_int_number={self.spg_int_symbol}, "
                f"crystal_system={self.crys_system})\n" + repr(self.lattice)
            )
        return f"{self.__class__.__name__}(spg_int_number={self.spg_int_symbol}, crystal_system={self.crys_system})\n"


def get_standardized_structure(structure: Structure) -> Structure:
    """
    Get standardized structure.

    Args:
        structure (Structure): Pymatgen Structure object.
    """
    species = [dict(site.species.as_dict()) for site in structure]
    sa = SpacegroupAnalyzer(structure)
    sd = sa.get_symmetry_dataset()
    std_lattice = sd["std_lattice"]
    mapping_to_primitive = sd["mapping_to_primitive"].tolist()
    mapping_to_primitive = np.array([mapping_to_primitive.index(i) for i in range(max(mapping_to_primitive) + 1)])
    std_species = np.array(species)[mapping_to_primitive][sd["std_mapping_to_primitive"]]
    std_frac_coords = sd["std_positions"]
    return Structure(lattice=std_lattice, species=std_species, coords=std_frac_coords).get_sorted_structure()



"""LAMMPS utility."""

from __future__ import annotations

import logging

import numpy as np
from pymatgen.core import Element, Lattice, Species, Structure
from pymatgen.core.operations import SymmOp

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

STRESS_FORMAT = {
    "VASP": ["xx", "yy", "zz", "xy", "yz", "xz"],
    "LAMMPS": ["xx", "yy", "zz", "xy", "xz", "yz"],
    "SNAP": ["xx", "yy", "zz", "yz", "xz", "xy"],
}


def check_structures_forces_stresses(
    structures: list[Structure],
    forces: list | None = None,
    stresses: list | None = None,
    stress_format: str = "VASP",
    return_none: bool = True,
):
    """
    Check structures, forces and stresses. The forces and stress are dependent
    on the lattice orientation. This function will rotate the structures
    and the corresponding forces and structures to lammps format
    [[ax, 0, 0],
    [bx, by, 0],
    [cx, cy, cz]].

    The lattice are formed by the row vectors.

    Args:
        structures (list): list of structures
        forces (list): list of force matrixs (m, 3)
        stresses (list): list of stress vectors
        stress_format (str): stress format, choose from
            "VASP", "LAMMPS", "SNAP"
        return_none (bool): whether to return list of None
            for forces and stresses

    Returns: structures [forces], [stresses]

    """
    new_structures = []
    new_forces = []
    new_stresses = []

    no_force = forces is None
    no_stress = stresses is None

    if forces is None:
        forces = [None] * len(structures)
    if stresses is None:
        stresses = [None] * len(structures)

    for i, s in enumerate(structures):
        # orthogonal structures do not need to rotate
        if s.lattice.is_orthogonal:
            new_structures.append(s)
            new_forces.append(forces[i])  # type: ignore
            new_stresses.append(stresses[i])  # type: ignore
            continue

        logger.info(f"Structure index {i} is rotated.")
        new_latt_matrix, symmop, rot_matrix = get_lammps_lattice_and_rotation(s, (0, 0, 0))
        coords = symmop.operate_multi(s.cart_coords)
        new_s = Structure(
            Lattice(new_latt_matrix),  # type: ignore
            s.species,  # type: ignore
            coords,  # type: ignore
            site_properties=s.site_properties,
            coords_are_cartesian=True,
        )
        new_structures.append(new_s)

        if not no_force:
            new_f = symmop.operate_multi(forces[i])
            new_forces.append(new_f)
        else:
            new_forces.append(None)

        if not no_stress:
            if np.size(stresses[i]) != 9:
                # voigt stress format
                stress_matrix = stress_list_to_matrix(stresses[i], stress_format)
                stress_matrix = rot_matrix.dot(stress_matrix).dot(rot_matrix.T)
                stress_matrix = stress_matrix_to_list(stress_matrix, stress_format)
            else:
                # 3x3 stress matrix
                stress_matrix = rot_matrix.dot(stresses[i]).dot(rot_matrix.T)
            # R \sigma R^T stress rotation
            new_stresses.append(stress_matrix)
        else:
            new_stresses.append(None)

    if return_none:
        return new_structures, new_forces, new_stresses

    out = [new_structures]
    if not no_force:
        out += [new_forces]
    if not no_stress:
        out += [new_stresses]
    if len(out) == 1:
        return out[0]
    return out


def stress_matrix_to_list(stress_matrix: np.ndarray, stress_format: str = "VASP") -> np.ndarray:
    """
    Stress matrix to list representation
    Args:
        stress_matrix (np.ndarray): stress matrix 3x3
        stress_format (str): stress list format
    Returns: list of float stress vector.
    """
    vasp_format = np.array(
        [
            stress_matrix[0, 0],
            stress_matrix[1, 1],
            stress_matrix[2, 2],
            stress_matrix[0, 1],
            stress_matrix[1, 2],
            stress_matrix[0, 2],
        ]
    )
    return stress_format_change(vasp_format, "VASP", stress_format)


def stress_list_to_matrix(stress: np.ndarray | list[float], stress_format: str = "VASP") -> np.ndarray:
    """
    Convert a length-6 stress list to stress matrix 3x3.

    Args:
        stress (list of float): list of stress
        stress_format (str): Supported formats are the follows
            VASP: xx, yy, zz, xy, yz, xz
            LAMMPS: xx, yy, zz, xy, zx, yz
            SNAP: xx, yy, zz, yz, xz, xy
    Returns: 3x3 stress matrix
    """
    s = stress_format_change(stress, from_format=stress_format, to_format="VASP")
    xx, yy, zz, xy, yz, xz = s
    return np.array([[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]])


def stress_format_change(stress: np.ndarray | list[float], from_format: str, to_format: str) -> np.ndarray:
    """
    Convert stress format from from_format to to_format
    Args:
        stress (list of float): length-6 stress vector
        from_format (str): choose from "VASP", "LAMMPS", "SNAP"
        to_format (str): choose from "VASP", "LAMMPS", "SNAP".

    Returns: list of float stress vector
    """
    from_order = STRESS_FORMAT[from_format.upper()]
    to_order = STRESS_FORMAT[to_format.upper()]
    mapping = [from_order.index(i) for i in to_order[3:]]
    return np.array([stress[0], stress[1], stress[2], *[stress[i] for i in mapping]])


def get_lammps_lattice_and_rotation(structure: Structure, origin=(0, 0, 0)) -> tuple[np.ndarray, SymmOp, np.ndarray]:
    """
    Transform structure to lammps compatible structure. The lattice and rotation
    matrix are returned.

    Args:
        structure (Structure): pymatgen structure
        origin (tuple): origin coordinates

    Returns: new lattice, rotation symmetry operator, rotation matrix

    """
    lattice = structure.lattice
    a, b, c = lattice.abc
    xlo, ylo, zlo = origin
    xhi = a + xlo
    m = lattice.matrix
    xy = np.dot(m[1], m[0] / a)
    yhi = np.sqrt(b**2 - xy**2) + ylo
    xz = np.dot(m[2], m[0] / a)
    yz = (np.dot(m[1], m[2]) - xy * xz) / (yhi - ylo)
    zhi = np.sqrt(c**2 - xz**2 - yz**2) + zlo
    # tilt = None if lattice.is_orthogonal else [xy, xz, yz]
    new_matrix = np.array([[xhi - xlo, 0, 0], [xy, yhi - ylo, 0], [xz, yz, zhi - zlo]])
    rot_matrix = np.linalg.solve(new_matrix, m)
    symmop = SymmOp.from_rotation_and_translation(rot_matrix, origin)
    return new_matrix, symmop, rot_matrix


def write_data_from_structure(
    structure: Structure,
    filename: str,
    ff_elements: list[str] | None = None,
    significant_figures: int = 6,
    origin: tuple = (0, 0, 0),
):
    """
    Write structure to lammps data file, this is to speed up
    pymatgen LammpsData.

    Args:a
        structure (Structure): pymatgen structure
        filename (str): filename
        ff_elements (list of str): elements to be considered
        significant_figures (int): significant figures of floats in output
        origin (tuple): origin coordinates
    """
    new_matrix, symmop, rot_matrix = get_lammps_lattice_and_rotation(structure=structure, origin=origin)
    lattice = structure.lattice
    xlo, ylo, zlo = origin
    xhi = new_matrix[0, 0] + xlo
    yhi = new_matrix[1, 1] + ylo
    zhi = new_matrix[2, 2] + zlo
    xy = new_matrix[1, 0]
    xz = new_matrix[2, 0]
    yz = new_matrix[2, 1]

    tilt = None if lattice.is_orthogonal else [xy, xz, yz]
    bounds = [[xlo, xhi], [ylo, yhi], [zlo, zhi]]

    lines = ["# Atom data generated by maml package"]

    elements_in_structure = [str(i.specie) for i in structure.sites]

    if ff_elements is None:
        ff_elements = list(set(elements_in_structure))
    else:
        if not set(elements_in_structure).issubset(set(ff_elements)):
            raise ValueError("Structure contains elements not in ff_elements")

    n_types = len(ff_elements)

    element_map = {i: j + 1 for j, i in enumerate(ff_elements)}
    # generate atom section
    lines.append(f"{len(structure)}    atoms\n")
    lines.append(f"{n_types}    atom types\n")

    ph = f"{{:.{significant_figures}f}}"

    for bound, d in zip(bounds, "xyz", strict=False):
        line = " ".join([ph.format(i) for i in bound] + [f"{d}{i}" for i in ["lo", "hi"]])
        lines.append(line)
    if tilt is not None:
        line = " ".join([ph.format(i) for i in tilt] + ["xy", "xz", "yz"])
        lines.append(line)

    lines.append("\nMasses\n")
    masses = [("{} " + ph).format(j, _get_atomic_mass(i)) for i, j in element_map.items()]
    lines.extend(masses)
    lines.append("\nAtoms\n")

    new_coords = symmop.operate_multi(structure.cart_coords)
    for i, (site, coords) in enumerate(zip(structure.sites, new_coords, strict=False)):
        charge = _get_charge(site.specie)
        line = "{} {} " + " ".join([ph] * 4)
        index = i + 1
        type = element_map[str(site.specie)]
        line = line.format(index, type, charge, *coords)
        lines.append(line)
    with open(filename, "w") as f:
        f.write("\n".join(lines))


def _get_atomic_mass(element_or_specie: str) -> float:  # type: ignore
    """
    Get atomic mass from element or specie string.

    Args:
        element_or_specie (str): specie or element string

    Returns: float mass

    """
    try:
        return Element(element_or_specie).atomic_mass  # type: ignore
    except Exception:
        return Species.from_str(element_or_specie).element.atomic_mass


def _get_charge(element_or_specie: str | Element | Species) -> float:  # type: ignore
    """
    Get charge from element or specie.

    Args:
        element_or_specie (str or Element or Species): element or specie

    Returns: charge float

    """
    if isinstance(element_or_specie, Species):
        return element_or_specie.oxi_state  # type: ignore
    if isinstance(element_or_specie, str):
        try:
            return Species.from_string(element_or_specie).oxi_state  # type: ignore
        except Exception:
            return 0.0
    return 0.0



"""Multi-layer perceptron models."""

from __future__ import annotations

from maml.base import BaseDescriber, KerasModel


def construct_mlp(
    input_dim: int,
    n_neurons: tuple = (64, 64),
    activation: str = "relu",
    n_targets: int = 1,
    is_classification: bool = False,
    optimizer: str = "adam",
    loss: str = "mse",
    compile_metrics: tuple = (),
):
    """
    Constructor for multi-layer perceptron models.

    Args:
        input_dim (int): input dimension, i.e., feature dimension
        n_neurons (tuple): list of hidden neuron sizes
        activation (str): activation function
        n_targets (int): number of targets
        is_classification (bool): whether the target is a classification problem
        optimizer (str): optimizer
        loss (str): loss function
        compile_metrics (tuple): metrics to evaluate during epochs
    """
    from tensorflow.keras.layers import Dense, Input
    from tensorflow.keras.models import Model

    inp = Input(shape=(input_dim,))
    out_ = inp
    for n_neuron in n_neurons:
        out_ = Dense(n_neuron, activation=activation)(out_)

    if is_classification:
        final_act: str | None = "sigmoid"
    else:
        final_act = None
    out = Dense(n_targets, activation=final_act)(out_)

    model = Model(inputs=inp, outputs=out)
    model.compile(optimizer, loss, metrics=compile_metrics)
    return model


class MLP(KerasModel):
    """This class implements the multi-layer perceptron models."""

    def __init__(
        self,
        input_dim: int | None = None,
        describer: BaseDescriber | None = None,
        n_neurons: tuple = (64, 64),
        activation: str = "relu",
        n_targets: int = 1,
        is_classification: bool = False,
        optimizer: str = "adam",
        loss: str = "mse",
        compile_metrics: tuple = (),
        **kwargs,
    ):
        """
        Constructor for multi-layer perceptron models.

        Args:
            input_dim (int): input dimension, i.e., feature dimension
            describer: a BaseDescriber featurizer.
            n_neurons: number of neurons in each layer of MLP.
            activation (str): activation function
            n_targets (int): number of targets
            is_classification (bool): whether the target is a classification problem
            optimizer (str): optimizer
            loss (str): loss function
            compile_metrics (tuple): metrics to evaluate during epochs
            **kwargs: pass to constructor of KerasModel.
        """
        input_dim = self.get_input_dim(describer, input_dim)
        if input_dim is None:
            raise ValueError("input_dim is not known and cannot be inferred")
        model = construct_mlp(
            input_dim=input_dim,
            n_neurons=n_neurons,
            activation=activation,
            n_targets=n_targets,
            is_classification=is_classification,
            optimizer=optimizer,
            loss=loss,
            compile_metrics=compile_metrics,
        )
        super().__init__(describer=describer, model=model, **kwargs)



"""neural network models."""

from __future__ import annotations

import json
import math
import os
from typing import TYPE_CHECKING

import joblib
import numpy as np

from maml.base import BaseDescriber, BaseModel, KerasModel

from ._keras_utils import deserialize_keras_activation, deserialize_keras_optimizer

if TYPE_CHECKING:
    from collections.abc import Sequence


def construct_atom_sets(
    input_dim: int | None = None,
    is_embedding: bool = True,
    n_neurons: Sequence[int] = (64, 64),
    n_neurons_final: Sequence[int] = (64, 64),
    n_targets: int = 1,
    activation: str = "relu",
    embedding_vcal: int = 95,
    embedding_dim: int = 32,
    symmetry_func: list[str] | str = "mean",
    optimizer: str = "adam",
    loss: str = "mse",
    compile_metrics: tuple = (),
    is_classification: bool = False,
    **symmetry_func_kwargs,
):
    r"""
    f(X) = \rho(\sum_{x \in X} \phi(x)), where X is a set.
    \phi is implemented as a neural network and \rho is a symmetry function.

    todo: implement attention mechanism

    Args:
        input_dim (int): input dimension, if None, then integer inputs + embedding are assumed.
        is_embedding (bool): whether the input should be embedded
        n_neurons (tuple): number of hidden-layer neurons before passing to symmetry function
        n_neurons_final (tuple): number of hidden-layer neurons after symmetry function
        n_targets (int): number of output targets
        activation (str): activation function
        embedding_vcal (int): int, embedding vocabulary
        embedding_dim (int): int, embedding dimension
        symmetry_func (str): symmetry function, choose from ['set2set', 'sum', 'mean',
            'max', 'min', 'prod']
        optimizer (str): optimizer for the models
        loss (str): loss function for the models
        compile_metrics (tuple): metrics for validation
        is_classification (bool): Whether to construct a classification model.
        symmetry_func_kwargs (dict): kwargs for symmetry function
    """
    from tensorflow.keras.layers import Concatenate, Dense, Embedding, Input
    from tensorflow.keras.models import Model

    if is_embedding and input_dim is not None:
        raise ValueError("When embedding is used, input dim needs to be None")

    if is_embedding:
        inp = Input(shape=(None,), dtype="int32", name="node_id")
        out_ = Embedding(embedding_vcal, embedding_dim)(inp)
    else:
        inp = Input(shape=(None, input_dim), dtype="float32", name="node_feature_input")
        out_ = inp

    weight_inputs = Input(shape=(None,), dtype="float32", name="weight_input")
    node_ids = Input(shape=(None,), dtype="int32", name="node_in_graph_id")

    # start neural networks \phi
    for n_neuron in n_neurons:
        out_ = Dense(n_neuron, activation=activation)(out_)

    # apply symmetry function \rho
    if isinstance(symmetry_func, str):
        symmetry_func = [symmetry_func]

    symmetry_layers = []
    for symm in symmetry_func:
        if symm == "set2set":
            from maml.models.dl._layers import WeightedSet2Set

            layer = WeightedSet2Set(**symmetry_func_kwargs)
        elif symm == "mean":
            from maml.models.dl._layers import WeightedAverageLayer

            alpha = symmetry_func_kwargs.pop("alpha", 1)
            layer = WeightedAverageLayer(alpha=alpha)
        else:
            raise ValueError("symmetry function not supported")
        symmetry_layers.append(layer)
    outs = [i([out_, weight_inputs, node_ids]) for i in symmetry_layers]

    out_ = Concatenate(axis=-1)(outs) if len(outs) > 1 else outs[0]

    # neural networks
    for n_neuron in n_neurons_final:
        out_ = Dense(n_neuron, activation=activation)(out_)

    if is_classification:
        final_act: str | None = "sigmoid"
    else:
        final_act = None
    out_ = Dense(n_targets, activation=final_act)(out_)
    model = Model(inputs=[inp, weight_inputs, node_ids], outputs=out_)
    model.compile(optimizer, loss, metrics=compile_metrics)
    return model


class AtomSets(KerasModel):
    r"""This class implements the DeepSets models."""

    def __init__(
        self,
        describer: BaseDescriber | None = None,
        input_dim: int | None = None,
        is_embedding: bool = True,
        n_neurons: Sequence[int] = (64, 64),
        n_neurons_final: Sequence[int] = (64, 64),
        n_targets: int = 1,
        activation: str = "relu",
        embedding_vcal: int = 95,
        embedding_dim: int = 32,
        symmetry_func: list[str] | str = "mean",
        optimizer: str = "adam",
        loss: str = "mse",
        compile_metrics: tuple = (),
        is_classification: bool = False,
        **symmetry_func_kwargs,
    ):
        """
        Args:
            describer (BaseDescriber): site describers
            input_dim (int): input dimension, if None, then integer inputs + embedding are assumed.
            is_embedding (bool): whether the input should be embedded
            n_neurons (tuple): number of hidden-layer neurons before passing to symmetry function
            n_neurons_final (tuple): number of hidden-layer neurons after symmetry function
            n_targets (int): number of output targets
            activation (str): activation function
            embedding_vcal (int): int, embedding vocabulary
            embedding_dim (int): int, embedding dimension
            symmetry_func (str): symmetry function, choose from ['set2set', 'sum', 'mean',
                'max', 'min', 'prod']
            optimizer (str): optimizer for the models
            loss (str): loss function for the models
            compile_metrics (tuple): metrics for validation
            is_classification (bool): Whether to construct a classification model.
            symmetry_func_kwargs (dict): kwargs for symmetry function.

        """
        input_dim = self.get_input_dim(describer, input_dim)

        optimizer = deserialize_keras_optimizer(optimizer)
        activation = deserialize_keras_activation(activation)

        model = construct_atom_sets(
            input_dim=input_dim,
            is_embedding=is_embedding,
            n_neurons=n_neurons,
            n_neurons_final=n_neurons_final,
            n_targets=n_targets,
            activation=activation,
            embedding_vcal=embedding_vcal,
            embedding_dim=embedding_dim,
            symmetry_func=symmetry_func,
            optimizer=optimizer,
            loss=loss,
            compile_metrics=compile_metrics,
            is_classification=is_classification,
            **symmetry_func_kwargs,
        )

        self.input_dim = input_dim
        self.is_embedding = is_embedding
        self.n_neurons = n_neurons
        self.n_neurons_final = n_neurons_final
        self.n_targets = n_targets
        self.activation = activation
        self.embedding_vcal = embedding_vcal
        self.embedding_dim = embedding_dim
        self.symmetry_func = symmetry_func
        self.optimizer = optimizer
        self.loss = loss
        self.compile_metrics = compile_metrics
        self.is_classification = is_classification
        self.symmetry_func_kwargs = symmetry_func_kwargs
        super().__init__(model=model, describer=describer)

    def _get_data_generator(self, features, targets, batch_size=128, is_shuffle=True):
        if features is None:
            return None
        from tensorflow.keras.utils import Sequence as KerasSequence

        def _generate_atom_indices(lengths):
            max_length = max(lengths)
            res = np.tile(np.arange(max_length)[None, :], (len(lengths), 1))
            res2 = np.tile(np.arange(len(lengths))[:, None], (1, max_length))
            return res2[res < np.array(lengths)[:, None]]

        is_embedding = self.is_embedding

        class _DataGenerator(KerasSequence):
            def __init__(
                self,
                features=features,
                targets=targets,
                batch_size=batch_size,
                is_shuffle=is_shuffle,
            ):
                if isinstance(features[0], list) and len(features[0]) == 2:
                    self.features = [i[0] for i in features]
                    self.weights = [i[1] for i in features]
                else:
                    self.features = features
                    self.weights = [np.ones(shape=(len(i),)) for i in features]
                self.targets = targets
                self.batch_size = batch_size
                self.is_shuffle = is_shuffle

            def __len__(self):
                return math.ceil(len(self.features) / self.batch_size)

            def __getitem__(self, idx):
                features_temp = self.features[idx * self.batch_size : (idx + 1) * self.batch_size]
                batch_x = np.concatenate(features_temp, axis=0)
                if is_embedding:
                    batch_x = batch_x[..., 0]
                lengths = [len(i) for i in features_temp]
                f_index = _generate_atom_indices(lengths)
                batch_weights = self.weights[idx * self.batch_size : (idx + 1) * self.batch_size]
                batch_y = self.targets[idx * self.batch_size : (idx + 1) * self.batch_size]
                return (
                    np.array(batch_x)[None, ...],
                    np.concatenate(batch_weights)[None, :],
                    f_index[None, :],
                ), np.array(batch_y)[None, :]

            def on_epoch_end(self):
                """Codes executed at the end of each epoch."""
                if self.is_shuffle:
                    indices = list(range(len(self.features)))
                    np.random.shuffle(indices)
                    self.features = [self.features[i] for i in indices]
                    self.weights = [self.weights[i] for i in indices]
                    self.targets = [self.targets[i] for i in indices]

        return _DataGenerator()

    def save(self, dirname: str):
        """Save the models and describers.

        Arguments:
            dirname (str): dirname for save
        """
        import tensorflow as tf

        if not os.path.isdir(dirname):
            os.mkdir(dirname)

        joblib.dump(self.describer, os.path.join(dirname, "describers.sav"))

        def _to_json_type(d):
            if isinstance(d, dict):
                return {i: _to_json_type(j) for i, j in d.items()}
            if isinstance(d, np.float32):
                return float(d)
            if isinstance(d, np.int32):
                return int(d)
            if isinstance(d, list | tuple):
                return [_to_json_type(i) for i in d]
            return d

        with open(os.path.join(dirname, "config.json"), "w") as f:
            json.dump(
                {
                    "input_dim": self.input_dim,
                    "is_embedding": self.is_embedding,
                    "n_neurons": self.n_neurons,
                    "n_neurons_final": self.n_neurons_final,
                    "n_targets": self.n_targets,
                    "activation": _to_json_type(tf.keras.activations.serialize(self.activation)),
                    "embedding_vcal": self.embedding_vcal,
                    "embedding_dim": self.embedding_dim,
                    "symmetry_func": self.symmetry_func,
                    "optimizer": _to_json_type(tf.keras.optimizers.serialize(self.optimizer)),
                    "loss": self.loss,
                    "compile_metrics": self.compile_metrics,
                    "is_classification": self.is_classification,
                    "symmetry_func_kwargs": self.symmetry_func_kwargs,
                },
                f,
            )
        self.model.save(os.path.join(dirname, "model_weights.hdf5"))

    @classmethod
    def from_dir(cls, dirname: str):
        """
        Load the models from file
        Args:
            dirname (str): directory name
        Returns: object instance.
        """
        with open(os.path.join(dirname, "config.json")) as f:
            kwarg_dict = json.load(f)

        symmetry_kwargs = kwarg_dict.pop("symmetry_func_kwargs")
        kwarg_dict.update(**symmetry_kwargs)

        describer = joblib.load(os.path.join(dirname, "describers.sav"))
        instance = cls(describer=describer, **kwarg_dict)
        instance.model.load_weights(os.path.join(dirname, "model_weights.hdf5"))
        return instance

    def fit(
        self,
        features: list | np.ndarray,
        targets: list | np.ndarray | None = None,
        val_features: list | np.ndarray | None = None,
        val_targets: list | np.ndarray | None = None,
        **kwargs,
    ) -> BaseModel:
        """
        Args:
            features (list or np.ndarray): Numerical input feature list or
                numpy array with dim (m, n) where m is the number of data and
                n is the feature dimension.
            targets (list or np.ndarray): Numerical output target list, or
                numpy array with dim (m, ).
            val_features (list or np.ndarray): validation features
            val_targets (list or np.ndarray): validation targets.
            **kwargs: pass to the fit method of BaseModel.

        """
        batch_size = kwargs.pop("batch_size", 128)
        is_shuffle = kwargs.pop("is_shuffle", True)
        train_generator = self._get_data_generator(features, targets, batch_size=batch_size, is_shuffle=is_shuffle)
        val_generator = self._get_data_generator(
            val_features, val_targets, batch_size=batch_size, is_shuffle=is_shuffle
        )
        return self.model.fit(train_generator, validation_data=val_generator, **kwargs)  # type: ignore

    def _predict(self, features: np.ndarray, **kwargs) -> np.ndarray:
        """
        Predict the values given a set of inputs based on fitted models.

        Args:
            features (np.ndarray): array-like input features.
            **kwargs: pass to the _get_data_generator method.

        Returns:
            List of output objects.
        """
        predict_generator = self._get_data_generator(features, [0] * len(features), is_shuffle=False, **kwargs)
        predicted = []
        for batch in predict_generator:
            predicted.append(self.model.predict(batch[0])[0])  # type: ignore
        return np.concatenate(predicted, axis=0)

    def evaluate(self, eval_objs, eval_targets, is_feature: bool = False, batch_size: int = 16):
        """
        Evaluate objs, targets.

        Args:
            eval_objs (list): objs for evaluation
            eval_targets (list): target list for the corresponding objects
            is_feature (bool): whether x is feature matrix
            batch_size (int): evaluation batch size
        """
        eval_features = eval_objs if is_feature else self.describer.transform(eval_objs)
        eval_generator = self._get_data_generator(eval_features, eval_targets, is_shuffle=False, batch_size=batch_size)
        return self.model.evaluate(eval_generator)



"""Module implements the describers for GB entry."""

from __future__ import annotations

from functools import reduce
from math import gcd

import numpy as np
import pandas as pd
from monty.json import MSONable
from pymatgen.analysis.gb.grain import GrainBoundary
from pymatgen.analysis.local_env import (
    BrunnerNN_real,
    BrunnerNN_reciprocal,
    BrunnerNN_relative,
    CovalentBondNN,
    Critic2NN,
    CrystalNN,
    CutOffDictNN,
    EconNN,
    JmolNN,
    MinimumDistanceNN,
    MinimumOKeeffeNN,
    MinimumVIRENN,
    NearNeighbors,
    OpenBabelNN,
    VoronoiNN,
)
from pymatgen.core import Element, Structure
from pymatgen.ext.matproj import MPRester, MPRestError
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

import maml.apps.gbe.presetfeatures as preset
from maml.apps.gbe.utils import load_b0_dict, load_mean_delta_bl_dict
from maml.base import BaseDescriber


def convert_hcp_direction(rotation_axis: list, lat_type: str) -> np.ndarray:
    """
    Four index notion to three index notion for hcp and rhombohedral axis
    Args:
        rotation_axis (list): four index notion of axis
        lat_type(str): the
    Returns:
        rotation axis in three index notion.
    """
    u1 = rotation_axis[0]
    v1 = rotation_axis[1]
    w1 = rotation_axis[3]
    if lat_type.lower().startswith("h"):
        u = 2 * u1 + v1
        v = 2 * v1 + u1
        w = w1
        rotation_axis = [u, v, w]
    elif lat_type.lower().startswith("r"):
        u = 2 * u1 + v1 + w1
        v = v1 + w1 - u1
        w = w1 - 2 * v1 - u1
        rotation_axis = [u, v, w]

    # make sure gcd(rotation_axis)==1
    if reduce(gcd, rotation_axis) != 1:
        rotation_axis = [round(x / reduce(gcd, rotation_axis)) for x in rotation_axis]
    return np.array(rotation_axis)


def convert_hcp_plane(plane: list) -> np.ndarray:
    """
    Four index notion to three index notion for hcp and rhombohedral plane
    Args:
        plane (list):  four index notion.

    Returns:
        three index notion of plane

    """
    u1 = plane[0]
    v1 = plane[1]
    w1 = plane[3]
    plane = [u1, v1, w1]
    if reduce(gcd, plane) != 1:
        index = reduce(gcd, plane)
        plane = [round(x / index) for x in plane]
    return np.array(plane)


class GBDescriber(BaseDescriber):
    """
    The describers that describes the grain boundary db entry
    with selected structural and elemental features.
    """

    def __init__(
        self,
        structural_features: list | None = None,
        elemental_features: list | None = None,
        **kwargs,
    ):
        """

        Args:
            structural_features (list): list of structural features
            elemental_features (list): list of elemental features
            **kwargs (dict): parameters for BaseDescriber.
        """
        if not elemental_features:
            elemental_features = [
                preset.e_coh,
                preset.G,
                preset.a0,
                preset.ar,
                preset.mean_delta_bl,
                preset.mean_bl,
            ]
        if not structural_features:
            structural_features = [
                preset.d_gb,
                preset.d_rot,
                preset.sin_theta,
                preset.cos_theta,
            ]
        self.elem_features = elemental_features
        self.struc_features = structural_features
        super().__init__(**kwargs)

    def transform_one(
        self,
        db_entry: dict,
        inc_target: bool = True,
        inc_bulk_ref: bool = True,
        mp_api: str | None = None,
    ) -> pd.DataFrame:
        """
        Describe gb with selected structural and elemental features
        Args:
            db_entry (dict): entry from surfacedb.All_GB_properties_parallel_copy
            inc_target (bool): whether to include target in the df, default: False
            inc_bulk_ref (bool): whether to generate bulk reference
                bulk reference: i.e. the entry of the origin bulk of the GB,
                                the rotation angle (theta) = 0, gb_energy = 0
            mp_api (str): MP api key.


        Returns: pd.DataFrame of processed data, columns are the feature labels
                 The task_id is included and serve as the unique id of the data

        """
        structural = get_structural_feature(db_entry=db_entry, features=self.struc_features)
        elemental = get_elemental_feature(db_entry=db_entry, features=self.elem_features, mp_api=mp_api)
        df = pd.concat([structural, elemental], axis=1, join="inner")
        df["task_id"] = db_entry["task_id"]
        if inc_target:
            df[preset.e_gb.str_name] = db_entry["gb_energy"]
        if inc_bulk_ref:
            df_bulk = self.generate_bulk_ref(df, inc_target)
            return pd.concat([df, df_bulk], axis=0)
        return df

    def generate_bulk_ref(self, gb_df: pd.DataFrame, inc_target: bool = True) -> pd.DataFrame:
        """
        Generate the bulk reference for given gb entry
        Args:
            gb_df (pd.DataFrame): data for gb
            inc_target (bool) : whether to include target.

        Returns:
            bulk_df (pd.DataFrame): data for bulk

        """
        new_df = gb_df.copy()
        new_df[preset.sin_theta.str_name] = np.sin(0)
        new_df[preset.cos_theta.str_name] = np.cos(0)
        if preset.mean_delta_bl.str_name in new_df:
            new_df[preset.mean_delta_bl.str_name] = 0
        if inc_target:
            new_df[preset.e_gb.str_name] = 0
        return new_df


def get_structural_feature(db_entry: dict, features: list | None = None) -> pd.DataFrame:
    """
    The structural features:
    d_gb: interplanal distance of the gb_plane
    d_rot: interplanal distance of the gb_plane
        w/ smallest integer index ) normal to rotation axis
    theta: rotation angle (sin and cos).

    Args:
        db_entry (Dict): db entry
        features (List): list of features.


    Returns:
        pd.DataFrame of structural features

    """
    if features is None:
        features = [preset.d_gb, preset.d_rot, preset.sin_theta, preset.cos_theta]
    if isinstance(db_entry["bulk_conv"], dict):
        bulk_conv = Structure.from_dict(db_entry["bulk_conv"])
    else:
        bulk_conv = db_entry["bulk_conv"]
    sg = SpacegroupAnalyzer(bulk_conv)
    gb_plane = np.array(db_entry["gb_plane"])
    rotation_axis = db_entry["rotation_axis"]
    if gb_plane.shape[0] == 4:
        gb_plane = convert_hcp_plane(list(gb_plane))
        rotation_axis = convert_hcp_direction(rotation_axis, lat_type=sg.get_crystal_system())
    d_gb = bulk_conv.lattice.d_hkl(gb_plane)  # type: ignore[arg-type]
    d_rot = bulk_conv.lattice.d_hkl(rotation_axis)
    theta = db_entry["rotation_angle"]
    sin_theta = np.sin(theta * np.pi / 180)
    cos_theta = np.cos(theta * np.pi / 180)
    fvalues = []
    for f in features:
        if f == preset.sin_theta:
            fvalues.append(np.sin(theta * np.pi / 180))
        elif f == preset.cos_theta:
            fvalues.append(np.cos(theta * np.pi / 180))
        else:
            fvalues.append(locals()[f"{f.str_name}"])
    return pd.DataFrame([fvalues], columns=[f.str_name for f in features])


def get_elemental_feature(
    db_entry: dict,
    loc_algo: str = "crystalnn",
    features: list | None = None,
    mp_api: str | None = None,
) -> pd.DataFrame:
    """
    Function to get the elemental features
    Args:
        db_entry (Dict): db entry
        loc_algo(str): algorithm to determine local env.
                    Options: see GBBond.NNDict.keys()
        features(List): list of feature names
                e_coh: cohesive energy
                G: G_vrh shear modulus
                a0: bulk lattice parameter a
                ar: atomic radius
                b0: the bond length of the metal bulk
                mean_delta_bl: the mean bond length difference
                    between GB and the bulk
                bl: the mean bond length in GB
        mp_api(str): api key to MP.

    Returns:
        pd.DataFrame of elemental features
    """
    if features is None:
        features = [
            preset.e_coh,
            preset.G,
            preset.a0,
            preset.ar,
            preset.mean_delta_bl,
            preset.bdensity,
            preset.CLTE,
            preset.hb,
        ]
    if preset.mean_delta_bl in features or preset.mean_bl in features:
        mdbl = load_mean_delta_bl_dict(loc_algo=loc_algo)
    f_dict = {}
    if mp_api:
        rester = MPRester(mp_api)
    else:
        raise MPRestError("Please provide API key to access Materials Project")
    bulk = rester.materials.search(material_ids=[db_entry["material_id"]])[0]  # type: ignore[attr-defined]
    bulk_s = rester.get_structure_by_material_id(db_entry["material_id"])  # type: ignore[attr-defined]
    if bulk:
        f_dict[preset.bdensity.str_name] = bulk[0]["density"]
        f_dict[preset.G.str_name] = bulk[0]["elasticity"]["G_VRH"]
    el = Element(db_entry["pretty_formula"])
    f_dict[preset.ar.str_name] = el.atomic_radius
    f_dict[preset.a0.str_name] = bulk_s.lattice.a
    f_dict[preset.e_coh.str_name] = rester.get_cohesive_energy(db_entry["material_id"])  # type: ignore[attr-defined]
    f_dict[preset.hb.str_name] = el.brinell_hardness
    f_dict[preset.CLTE.str_name] = el.coefficient_of_linear_thermal_expansion

    def get_mean_delta_bl(db_entry):
        return mdbl[str(db_entry["task_id"])]

    def get_bl(db_entry):
        b0_dict = load_b0_dict()
        b0 = b0_dict[db_entry["pretty_formula"]]
        return mdbl[str(db_entry["task_id"])] + b0

    fvalues = []
    for f in features:
        if f == preset.mean_delta_bl:
            # func = 'get_{}'.format(f.str_name).lower()
            # fvalues.append(locals()[func](db_entry))
            fvalues.append(get_mean_delta_bl(db_entry))
        elif f == preset.mean_bl:
            fvalues.append(get_bl(db_entry))
        else:
            fvalues.append(f_dict[f.str_name])
    return pd.DataFrame([fvalues], columns=[f.str_name for f in features])


class GBBond(MSONable):
    """
    This class describes the GB Bonding environment using provided
    local environment algorithm.
    Available algorithms: GBBond.NNDict.keys(), default CrystalNN.
    """

    NNDict = {
        i.__name__.lower(): i
        for i in [
            NearNeighbors,
            VoronoiNN,
            JmolNN,
            MinimumDistanceNN,
            OpenBabelNN,
            CovalentBondNN,
            MinimumVIRENN,
            MinimumOKeeffeNN,
            BrunnerNN_reciprocal,
            BrunnerNN_real,
            BrunnerNN_relative,
            EconNN,
            CrystalNN,
            CutOffDictNN,
            Critic2NN,
        ]
    }

    def __init__(
        self,
        gb: GrainBoundary,
        loc_algo: str = "crystalnn",
        bond_mat: np.ndarray | None = None,
        **kwargs,
    ):
        """

        Args:
            gb (GrainBoundary): the GrainBoundary Object
            loc_algo (str): algorithm to determine local env.
                    See options: GBBond.NNDict.keys()
                    Default: crystalnn
            bond_mat (np.ndarray): optional.
            **kwargs: pass to loc_algo.
        """
        self.loc_algo = self.NNDict[loc_algo](**kwargs)
        self.gb = gb
        if isinstance(bond_mat, np.ndarray):
            self.bond_mat = bond_mat
        else:
            self.bond_mat = self._get_bond_mat(gb)

    def _get_bond_mat(self, gb: GrainBoundary) -> np.ndarray:
        """

        Args:
            gb (GrainBoundary): the grain boundary structure object.

        Returns:
            bond matrix
            matrix of bond lengths
            1. bm[i][j] = bond length between atom i&j
            if i & j is bonded (determined by the loc_algo)
            2. bm[i][j] = bm[j][i]
            3. If not bonded, the bm[i][j] = np.nan

        """
        if isinstance(gb, dict):
            gb = GrainBoundary.from_dict(gb)
        bond_mat = np.zeros((gb.num_sites, gb.num_sites))
        all_nn_info = self.loc_algo.get_all_nn_info(gb)
        for i, _ in enumerate(gb.sites):
            # compute the all_nn_info first then call
            # the protected function is faster for multiple computation
            nn = self.loc_algo._get_nn_shell_info(structure=gb, site_idx=i, shell=1, all_nn_info=all_nn_info)
            for b in nn:
                bl = gb.get_distance(i, b["site_index"])
                bond_mat[i][b["site_index"]] = bl
        return bond_mat

    def get_breakbond_ratio(self, gb: GrainBoundary, b0: float, return_count: bool = False):
        """
        Get the breakbond ratio, i.e the ratio of shorter bonds vs longer bonds
        compared to the bulk bond length (b0)
        The algo to find the bonds can vary
        Meme: if use get_neighbors, require a hard set cutoff, which adds to
            arbitrariness.

        Args:
            gb (GrainBoundary): a GrainBoundary object
            b0 (float): cutoff to determine short vs. long bonds,
                        default the bulk bond length
            return_count(bool): whether to return count of

        Returns:
            ratio (float): shorter_bond / longer_bonds
            if return_count:
            shorter_bond: # of short bonds
            longer_bond: # of long bonds
        """
        if not self.bond_mat:
            self._get_bond_mat(gb)
        long_bond = (self.bond_mat > b0).sum()
        short_bond = (self.bond_mat < b0).sum()
        ratio = short_bond / long_bond if long_bond else 0
        if return_count:
            return ratio, short_bond, long_bond
        return ratio

    @property
    def min_bl(self) -> float:
        """The minimum bond length."""
        return self.bond_mat[self.bond_mat > 0].min()

    @property
    def max_bl(self) -> float:
        """The maximum bond length."""
        return self.bond_mat.max()

    @property
    def bond_matrix(self) -> np.ndarray:
        """The (padded) bond matrix."""
        return self.bond_mat

    def get_mean_bl_chg(self, b0: float) -> float:
        """
        Function to calculate the mean bond length difference between GB and the bulk
        Args:
            b0 (float): the bond length in bulk.

        Returns: the mean_bl_chg

        """
        return (self.bond_mat[self.bond_mat > 0] - b0).mean()

    def as_dict(self) -> dict:
        """
        Dict representation of the GBond class.

        Returns:
            dict of {"loc_algo": str, "bond_mat": bond matrix}
        """
        return_dict = super().as_dict()
        return_dict.update(
            {
                "loc_algo": self.loc_algo.__class__.__name__.lower(),
                "bond_mat": self.bond_mat,
                # "min_bl": self.min_bl,
                # "max_bl": self.max_bl
            }
        )
        return return_dict

    @classmethod
    def from_dict(cls, d):
        """

        Args:
            d (dict): Dict representation.

        Returns:
            GBBond

        """
        return cls(gb=d["gb"], loc_algo=d["loc_algo"], bond_mat=d["bond_mat"])



"""Deep learning module."""

from __future__ import annotations

from ._atomsets import AtomSets
from ._layers import WeightedAverageLayer, WeightedSet2Set
from ._mlp import MLP

__all__ = ["MLP", "AtomSets", "WeightedAverageLayer", "WeightedSet2Set"]



"""Define several typing for convenient use."""

from __future__ import annotations

from collections.abc import Callable

import numpy as np
from pymatgen.core import Molecule, Structure

OptStrOrCallable = str | Callable | None
StructureOrMolecule = Structure | Molecule
VectorLike = list[float] | np.ndarray



"""MAML models."""

from __future__ import annotations

from maml.base import KerasModel, SKLModel

from .dl import MLP, AtomSets, WeightedAverageLayer, WeightedSet2Set

__all__ = ["MLP", "AtomSets", "KerasModel", "SKLModel", "WeightedAverageLayer", "WeightedSet2Set"]



"""
This module implements more robust optimization
using the cvxpy package.
"""

from __future__ import annotations

import numpy as np
from monty.dev import requires
from scipy.linalg import lstsq

from maml.apps.symbolic._selectors import BaseSelector

try:
    import cvxpy as cp

    Expression = cp.expressions.expression.Expression
except ImportError:
    cp = None
    Expression = "Expression"


class BaseSelectorCP(BaseSelector):
    """Base selector using cvxpy (CP)."""

    @requires(cp is not None, "cvxpy is not present.")
    def __init__(self, coef_thres: float = 1e-6, method: str = "ECOS"):
        """
        Base selector
        Args:
            coef_thres (float): threshold to discard certain coefficients
            method (str): solver for cvxpy problem. The default is ECOS.
        """
        super().__init__(coef_thres=coef_thres, method=method)

    # pylint: disable=E1128
    def select(self, x: np.ndarray, y: np.ndarray, options: dict | None = None) -> np.ndarray | None:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): kwargs for cp.Problem.solve
        Returns: list of int indices.
        """
        options = options or {}
        n, p = x.shape
        beta = cp.Variable(p)
        objective = cp.Minimize(self.construct_loss(x, y, beta))
        constraints = self.construct_constraints(x, y, beta)
        prob = cp.Problem(objective, constraints)
        prob.solve(solver=self.method, **options)
        self.coef_ = beta.value
        self.indices = np.where(np.abs(beta.value) > self.coef_thres)[0]
        self.coef_[np.where(np.abs(self.coef_) <= self.coef_thres)[0]] = 0.0  # type: ignore
        return self.indices

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: cp.Variable | None = None
    ) -> list[Expression] | None:  # type: ignore
        """
        Get constraints dictionary from data, e.g.,
        {"func": lambda beta: fun(x, y, beta), "type": "ineq"}.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta: (np.ndarray): target variable for optimization
        Returns: dict of constraints.
        """
        return None

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: cp.Variable) -> Expression:  # type: ignore
        """
        Get loss function from data and tentative coefficients beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: loss value.
        """
        raise NotImplementedError


class DantzigSelectorCP(BaseSelectorCP):
    """
    Equation 11 in
    https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf
    and reference in https://projecteuclid.org/download/pdfview_1/euclid.aos/1201012958.
    """

    @requires(cp is not None, "cvxpy is not present.")
    def __init__(self, lambd, sigma=1.0, **kwargs):
        """
        Dantzig selector
        Args:
            lamb: tunable parameter
            sigma: standard deviation of the error.
        """
        self.lambd = lambd
        self.sigma = sigma
        super().__init__(**kwargs)

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: cp.Variable) -> Expression:  # type: ignore
        """
        L1 loss
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (cp.Variable): dimension N vector for optimization
        Returns: loss expression.
        """
        return cp.norm1(beta)

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: cp.Variable | None = None
    ) -> list[Expression] | None:  # type: ignore
        """
        Dantzig selector constraints
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (cp.Variable): dimension N vector for optimization
        Returns: List of constraints.
        """
        return [cp.norm_inf(x.T @ (y - x @ beta)) <= self.lambd * self.sigma]


class PenalizedLeastSquaresCP(BaseSelectorCP):
    """
    Penalized least squares. In addition to minimizing the sum of squares loss,
    it adds an additional penalty to the coefficients.
    """

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: cp.Variable) -> Expression:  # type: ignore
        """
        L1 loss
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (cp.Variable): dimension N vector for optimization
        Returns: loss expression.
        """
        n = x.shape[0]
        return 1.0 / (2 * n) * cp.sum_squares(y - x @ beta) + self.penalty(beta, x=x, y=y)

    def penalty(
        self,
        beta: cp.Variable,
        x: np.ndarray | None = None,
        y: np.ndarray | None = None,
    ) -> Expression | float:  # type: ignore
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value.
        """
        return 0.0


class LassoCP(PenalizedLeastSquaresCP):
    """Simple Lasso regression."""

    @requires(cp is not None, "cvxpy not installed")
    def __init__(self, lambd, **kwargs):
        """
        Lasso regression with lambda * norm_1(beta) as penalty.

        Args:
            lambd (float): weights for the penalty.
            **kwargs: pass to parent class PenalizedLeastSquaresCP.
        """
        self.lambd = lambd
        super().__init__(**kwargs)

    def penalty(
        self,
        beta: cp.Variable,
        x: np.ndarray | None = None,
        y: np.ndarray | None = None,
    ) -> Expression | float:  # type: ignore
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value.
        """
        beta_abs = cp.norm1(beta)
        return self.lambd * beta_abs


class AdaptiveLassoCP(PenalizedLeastSquaresCP):
    """
    Adaptive lasso regression using OLS coefficients
    as the root-n estimator coefficients.
    """

    @requires(cp is not None, "cvxpy not installed")
    def __init__(self, lambd, gamma, **kwargs):
        """
        Adaptive lasso regression.

        Args:
            lambd (float or list of floats): the regularization parameter lambda.
            gamma (float): exponential for hat(beta)
            **kwargs: pass to parent class PenalizedLeastSquaresCP.
        """
        self.lambd = lambd
        self.gamma = gamma
        self.w = 1
        super().__init__(**kwargs)

    def select(self, x: np.ndarray, y: np.ndarray, options: dict | None = None) -> np.ndarray | None:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): options in the cp.Problem.solve
        Returns: array int indices.
        """
        self.w = self.get_w(x, y)
        return super().select(x, y, options)

    def get_w(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Get adaptive weights from data
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: coefficients array.
        """
        beta_hat = lstsq(x, y)[0]
        return 1.0 / np.abs(beta_hat) ** self.gamma

    def penalty(
        self,
        beta: cp.Variable,
        x: np.ndarray | None = None,
        y: np.ndarray | None = None,
    ) -> Expression | float:  # type: ignore
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value.
        """
        return cp.sum(self.lambd * cp.multiply(self.w, cp.abs(beta)))



"""Materials utils."""

from __future__ import annotations

from pymatgen.core import Composition, Molecule, Structure


def to_composition(obj: Composition | Molecule | Structure | str) -> Composition:
    """
    Convert str/structure or composition to compositions.

    Args:
        obj (str/structure/composition): object to convert

    Returns:
        Composition object
    """
    if isinstance(obj, str):
        return Composition(obj)
    if isinstance(obj, Structure | Molecule):
        return obj.composition
    return obj



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides basic LAMMPS calculator classes."""
from __future__ import annotations

import abc
import io
import itertools
import logging
import os
import subprocess
from shutil import which

import numpy as np
from monty.tempfile import ScratchDir
from pymatgen.core import Element, Lattice, Structure
from pymatgen.io.lammps.data import LammpsData

from maml.apps.pes._base import Potential
from maml.utils import get_lammps_lattice_and_rotation, stress_list_to_matrix, stress_matrix_to_list

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_default_lmp_exe():
    """
    Get lammps executable
    Returns: Lammps executable name.
    """
    for lmp_exe in ["lmp_serial", "lmp_mpi", "lmp_g++_serial", "lmp_g++_mpich", "lmp_intel_cpu_intelmpi"]:
        lmp_exe_detected = which(lmp_exe)
        if lmp_exe_detected is not None:
            logger.info(f"Setting Lammps executable to {lmp_exe_detected}")
            return lmp_exe_detected
    return None


def _pretty_input(lines):
    def prettify(line):
        return (
            line.split()[0].ljust(width) + " ".join(line.split()[1:])
            if not (len(line.split()) == 0 or line.strip().startswith("#"))
            else line
        )

    clean_lines = [line.strip("\n") for line in lines]
    commands = [line for line in clean_lines if len(line.strip()) > 0]
    keys = [c.split()[0] for c in commands if not c.split()[0].startswith("#")]
    width = max(len(k) for k in keys) + 4
    new_lines = map(prettify, clean_lines)
    return "\n".join(new_lines)


def _read_dump(file_name, dtype="float_"):
    with open(file_name) as f:
        lines = f.readlines()[9:]
    return np.loadtxt(io.StringIO("".join(lines)), dtype=dtype)


class LMPStaticCalculator:
    """
    Abstract class to perform static structure property calculation
    using LAMMPS.
    """

    _COMMON_CMDS = ["units metal", "atom_style charge", "box tilt large", "read_data data.static", "run 0"]

    allowed_kwargs = ["lmp_exe"]

    def __init__(self, **kwargs):
        """
        Initializer for lammps calculator
        Allowed keyword args are lmp_exe string.

        """
        lmp_exe = kwargs.pop("lmp_exe", None)
        if lmp_exe is None:
            lmp_exe = get_default_lmp_exe()
        if not which(lmp_exe):
            raise ValueError(f"lammps executable {lmp_exe} not found")
        self.LMP_EXE = lmp_exe
        for i, j in kwargs.items():
            if i not in self.allowed_kwargs:
                raise TypeError(f"{i!s} not in supported kwargs {self.allowed_kwargs!s}")
            setattr(self, i, j)

    @abc.abstractmethod
    def _setup(self):
        """Setup a calculation, writing input files, etc."""
        return

    @abc.abstractmethod
    def _sanity_check(self, structure):
        """Check if the structure is valid for this calculation."""
        return

    @abc.abstractmethod
    def _parse(self):
        """Parse results from dump files."""
        return

    def calculate(self, structures):
        """
        Perform the calculation on a series of structures.

        Args:
            structures: Input structures in a list.

        Returns:
            List of computed data corresponding to each structure,
            varies with different subclasses.

        """
        for struct in structures:
            assert self._sanity_check(struct) is True, "Incompatible structure found"
        ff_elements = None
        if hasattr(self, "element_profile"):
            element_profile = self.element_profile
            ff_elements = element_profile.keys()
        if hasattr(self, "ff_settings"):
            ff_settings = self.ff_settings
            if hasattr(ff_settings, "elements"):
                ff_elements = ff_settings.elements

        with ScratchDir("."):
            input_file = self._setup()
            data = []
            for struct in structures:
                struct.remove_oxidation_states()
                ld = LammpsData.from_structure(struct, ff_elements, atom_style="charge")
                ld.write_file("data.static")
                with subprocess.Popen([self.LMP_EXE, "-in", input_file], stdout=subprocess.PIPE) as p:
                    stdout = p.communicate()[0]
                    rc = p.returncode
                if rc != 0:
                    error_msg = f"LAMMPS exited with return code {rc}"
                    msg = stdout.decode("utf-8").split("\n")[:-1]
                    try:
                        error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                        error_msg += ", ".join(msg[error_line:])
                    except Exception:
                        error_msg += msg[-1]
                    raise RuntimeError(error_msg)
                results = self._parse()
                data.append(results)
        return data

    def set_lmp_exe(self, lmp_exe: str) -> None:
        """
        Set lammps executable for the instance.

        Args:
            lmp_exe (str): lammps executable path
        Returns:
        """
        self.LMP_EXE = lmp_exe


class EnergyForceStress(LMPStaticCalculator):
    """Calculate energy, forces and virial stress of structures."""

    def __init__(self, ff_settings, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for LAMMPS
                calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            **kwargs: Passthrough to init.
        """
        self.ff_settings = ff_settings
        super().__init__(**kwargs)

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "efs")
        with open(os.path.join(template_dir, "in.efs")) as f:
            input_template = f.read()

        input_file = "in.efs"

        ff_settings = self.ff_settings.write_param() if isinstance(self.ff_settings, Potential) else self.ff_settings

        with open(input_file, "w") as f:
            f.write(input_template.format(ff_settings="\n".join(ff_settings)))
        return input_file

    def _sanity_check(self, structure):
        return True

    def calculate(self, structures):
        """
        Calculate the energy, forces and stresses of structures.
        Proper rotation of the results are applied when the structure
        is triclinic.

        Args:
            structures (list): a list of structures

        Returns: list of (energy, forces, stresses) tuple

        """
        results = super().calculate(structures=structures)
        final_results = []
        for res, s in zip(results, structures, strict=False):
            new_forces, new_stresses = self._rotate_force_stress(s, res[1], res[2])
            final_results.append((res[0], new_forces, new_stresses))
        return final_results

    @staticmethod
    def _rotate_force_stress(structure, forces, stresses):
        _, symmop, rot_matrix = get_lammps_lattice_and_rotation(structure)
        inv_rot_matrix = np.linalg.inv(rot_matrix)
        forces = forces.dot(inv_rot_matrix.T)
        stresses = stress_list_to_matrix(stresses, stress_format="LAMMPS")
        stresses = inv_rot_matrix.dot(stresses.dot(inv_rot_matrix.T))
        stresses = stress_matrix_to_list(stresses, stress_format="LAMMPS")
        return forces, stresses

    def _parse(self):
        energy = float(np.loadtxt("energy.txt"))
        force = _read_dump("force.dump")
        stress = np.loadtxt("stress.txt")
        return energy, force, stress


class SpectralNeighborAnalysis(LMPStaticCalculator):
    """
    Calculator for bispectrum components to characterize the local
    neighborhood of each atom in a general way.

    Usage:
        [(b, db, e)] = sna.calculate([Structure])
        b: 2d NumPy array with shape (N, n_bs) containing bispectrum
            coefficients, where N is the No. of atoms in structure and
            n_bs is the No. of bispectrum components.
        db: 2d NumPy array with shape (N, 3 * n_bs * n_elements)
            containing the first order derivatives of bispectrum
            coefficients with respect to atomic coordinates,
            where n_elements is the No. of elements in element_profile.
        e: 2d NumPy array with shape (N, 1) containing the element of
            each atom.

    """

    _CMDS = [
        "pair_style lj/cut 10",
        "pair_coeff * * 1 1",
        "compute sna all sna/atom ",
        "compute snad all snad/atom ",
        "compute snav all snav/atom ",
        "dump 1 all custom 1 dump.element element",
        "dump_modify 1 sort id",
        "dump 2 all custom 1 dump.sna c_sna[*]",
        "dump_modify 2 sort id",
        "dump 3 all custom 1 dump.snad c_snad[*]",
        "dump_modify 3 sort id",
        "dump 4 all custom 1 dump.snav c_snav[*]",
        "dump_modify 4 sort id",
    ]

    def __init__(self, rcutfac, twojmax, element_profile, quadratic=False, **kwargs):
        """
        For more details on the parameters, please refer to the
        official documentation of LAMMPS.

        Notes:
            Despite this calculator uses compute sna(d)/atom command
            (http://lammps.sandia.gov/doc/compute_sna_atom.html), the
            parameter definition is in consistent with pair_style snap
            document (http://lammps.sandia.gov/doc/pair_snap.html).

        Args:
            rcutfac (float): The rcutfac in bispectrum coefficient calculations.
                The cutoff radius between element i and j are rcutfac * (R_i + R_j)
                where R_i and R_j are cutoff set for element i and j.
            twojmax (int): Band limit for bispectrum components.
            element_profile (dict): Parameters (cutoff radius 'r' and
                weight 'w') related to each element, e.g.,
                {'Na': {'r': 4.5, 'w': 0.9},
                 'Cl': {'r': 4.8, 'w': 3.0}}
            quadratic (bool): Whether including quadratic terms.
                Default to False.
            **kwargs: Passthrough to init.
        """
        self.rcutfac = rcutfac
        self.twojmax = twojmax
        self.element_profile = element_profile
        self.quadratic = quadratic
        super().__init__(**kwargs)

    @staticmethod
    def get_bs_subscripts(twojmax):
        """
        Method to list the subscripts 2j1, 2j2, 2j of bispectrum
        components.

        Args:
            twojmax (int): Band limit for bispectrum components.

        Returns:
            List of all subscripts [2j1, 2j2, 2j].

        """
        subs = itertools.product(range(twojmax + 1), repeat=3)

        filters = [lambda x: x[0] >= x[1], lambda x: x[2] >= x[0]]
        j_filter = [lambda x: x[2] in range(x[0] - x[1], min(twojmax, x[0] + x[1]) + 1, 2)]
        filters.extend(j_filter)
        for f in filters:
            subs = filter(f, subs)
        return list(subs)

    @property
    def n_bs(self):
        """Returns No. of bispectrum components to be calculated."""
        return len(self.get_bs_subscripts(self.twojmax))

    def _setup(self):
        def add_args(line):
            return line + compute_args if line.startswith("compute") else line

        compute_args = f"1 0.99363 {self.twojmax} "
        el_in_seq = sorted(self.element_profile.keys(), key=lambda x: Element(x))
        cutoffs = [self.element_profile[e]["r"] * self.rcutfac for e in el_in_seq]
        weights = [self.element_profile[e]["w"] for e in el_in_seq]
        compute_args += " ".join([str(p) for p in cutoffs + weights])
        compute_args += f" rmin0 0 quadraticflag {int(self.quadratic)}"
        CMDS = list(map(add_args, self._CMDS))
        CMDS[2] += " bzeroflag 0"
        CMDS[3] += " bzeroflag 0"
        CMDS[4] += " bzeroflag 0"
        dump_modify = "dump_modify 1 element "
        dump_modify += " ".join(str(e) for e in el_in_seq)
        CMDS.append(dump_modify)
        ALL_CMDS = self._COMMON_CMDS[:]
        ALL_CMDS[-1:-1] = CMDS
        input_file = "in.sna"
        with open(input_file, "w") as f:
            f.write(_pretty_input(ALL_CMDS))
        return input_file

    def _sanity_check(self, structure):
        elements = set(structure.symbol_set)
        sna_elements = self.element_profile.keys()
        return elements.issubset(sna_elements)

    def _parse(self):
        element = np.atleast_1d(_read_dump("dump.element", "unicode"))
        b = np.atleast_2d(_read_dump("dump.sna"))
        db = np.atleast_2d(_read_dump("dump.snad"))
        vb = np.atleast_2d(_read_dump("dump.snav"))
        return b, db, vb, element


class ElasticConstant(LMPStaticCalculator):
    """Elastic constant calculator."""

    _RESTART_CONFIG = {
        "internal": {"write_command": "write_restart", "read_command": "read_restart", "restart_file": "restart.equil"},
        "external": {"write_command": "write_restart", "read_command": "read_restart", "restart_file": "restart.equil"},
    }

    def __init__(
        self,
        ff_settings,
        potential_type="external",
        deformation_size=1e-6,
        jiggle=1e-5,
        maxiter=400,
        maxeval=1000,
        full_matrix=False,
        **kwargs,
    ):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for LAMMPS
                calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            potential_type (str): 'internal' indicates the internal potentials
                installed in lammps, 'external' indicates the external potentials
                outside of lammps.
            deformation_size (float): Finite deformation size. Usually range from
                1e-2 to 1e-8, to confirm the results not depend on it.
            jiggle (float): The amount of random jiggle for atoms to
                prevent atoms from staying on saddle points.
            maxiter (float): The maximum number of iteration. Default to 400.
            maxeval (float): The maximum number of evaluation. Default to 1000.
            full_matrix (bool): If False, only c11, c12, c44 and bulk modulus are returned.
                If True, 6 x 6 elastic matrices in the Voigt notation are returned.
            **kwargs: Passthrough to init.
        """
        self.ff_settings = ff_settings
        self.write_command = self._RESTART_CONFIG[potential_type]["write_command"]
        self.read_command = self._RESTART_CONFIG[potential_type]["read_command"]
        self.restart_file = self._RESTART_CONFIG[potential_type]["restart_file"]
        self.deformation_size = deformation_size
        self.jiggle = jiggle
        self.maxiter = maxiter
        self.maxeval = maxeval
        self.full_matrix = full_matrix
        super().__init__(**kwargs)

    def _setup(
        self,
    ):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "elastic")

        with open(os.path.join(template_dir, "in.elastic")) as f:
            input_template = f.read()
        with open(os.path.join(template_dir, "init.template")) as f:
            init_template = f.read()
        with open(os.path.join(template_dir, "potential.template")) as f:
            potential_template = f.read()
        with open(os.path.join(template_dir, "displace.template")) as f:
            displace_template = f.read()

        input_file = "in.elastic"
        ff_settings = self.ff_settings.write_param() if isinstance(self.ff_settings, Potential) else self.ff_settings

        with open(input_file, "w") as f:
            f.write(input_template.format(write_restart=self.write_command, restart_file=self.restart_file))

        with open("init.mod", "w") as f:
            f.write(
                init_template.format(
                    deformation_size=self.deformation_size,
                    jiggle=self.jiggle,
                    maxiter=self.maxiter,
                    maxeval=self.maxeval,
                )
            )
        with open("potential.mod", "w") as f:
            f.write(potential_template.format(ff_settings="\n".join(ff_settings)))
        with open("displace.mod", "w") as f:
            f.write(displace_template.format(read_restart=self.read_command, restart_file=self.restart_file))
        return input_file

    def _sanity_check(self, structure):
        """Check if the structure is valid for this calculation."""
        return True

    def _parse(self):
        """Parse results from dump files."""
        if self.full_matrix:
            return np.loadtxt("voigt_tensor.txt")

        C11, C12, C44, bulkmodulus = np.loadtxt("elastic.txt")
        return C11, C12, C44, bulkmodulus


class NudgedElasticBand(LMPStaticCalculator):
    """NudgedElasticBand migration energy calculator."""

    def __init__(self, ff_settings, specie, lattice, alat, num_replicas=7, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
            num_replicas (int): Number of replicas to use.
            **kwargs: Passthrough to init.
        """
        self.ff_settings = ff_settings
        self.specie = specie
        self.lattice = lattice
        self.alat = alat
        self.num_replicas = num_replicas
        super().__init__(**kwargs)

    @staticmethod
    def get_unit_cell(specie, lattice, alat):
        """
        Get the unit cell from specie, lattice type and lattice constant.

        Args:
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
        """
        if lattice == "fcc":
            unit_cell = Structure.from_spacegroup(
                sg="Fm-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "bcc":
            unit_cell = Structure.from_spacegroup(
                sg="Im-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "diamond":
            unit_cell = Structure.from_spacegroup(
                sg="Fd-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        else:
            raise ValueError("Lattice type is invalid.")

        return unit_cell

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "neb")

        with open(os.path.join(template_dir, "in.relax")) as f:
            relax_template = f.read()
        with open(os.path.join(template_dir, "in.neb")) as f:
            neb_template = f.read()

        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=self.alat)
        lattice_calculator = LatticeConstant(ff_settings=self.ff_settings)
        a, _, _ = lattice_calculator.calculate([unit_cell])[0]
        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=a)

        if self.lattice == "fcc":
            start_idx, final_idx = 95, 49
            scale_factor = [3, 3, 3]
        elif self.lattice == "bcc":
            start_idx, final_idx = 40, 14
            scale_factor = [3, 3, 3]
        elif self.lattice == "diamond":
            start_idx, final_idx = 7, 15
            scale_factor = [2, 2, 2]
        else:
            raise ValueError("Lattice type is invalid.")

        super_cell = unit_cell * scale_factor
        super_cell_ld = LammpsData.from_structure(
            super_cell, ff_elements=self.ff_settings.elements, atom_style="charge"
        )
        super_cell_ld.write_file("data.supercell")

        with open("in.relax", "w") as f:
            f.write(
                relax_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    lattice=self.lattice,
                    alat=a,
                    specie=self.specie,
                    del_id=start_idx + 1,
                    relaxed_file="initial.relaxed",
                )
            )

        with subprocess.Popen([self.LMP_EXE, "-in", "in.relax"], stdout=subprocess.PIPE) as p:
            stdout = p.communicate()[0]
            rc = p.returncode
        if rc != 0:
            error_msg = f"LAMMPS exited with return code {rc}"
            msg = stdout.decode("utf-8").split("\n")[:-1]
            try:
                error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                error_msg += ", ".join(msg[error_line:])
            except Exception:
                error_msg += msg[-1]
            raise RuntimeError(error_msg)

        with open("in.relax", "w") as f:
            f.write(
                relax_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    lattice=self.lattice,
                    alat=a,
                    specie=self.specie,
                    del_id=final_idx + 1,
                    relaxed_file="final.relaxed",
                )
            )

        with subprocess.Popen([self.LMP_EXE, "-in", "in.relax"], stdout=subprocess.PIPE) as p:
            stdout = p.communicate()[0]
            rc = p.returncode
        if rc != 0:
            error_msg = f"LAMMPS exited with return code {rc}"
            msg = stdout.decode("utf-8").split("\n")[:-1]
            try:
                error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                error_msg += ", ".join(msg[error_line:])
            except Exception:
                error_msg += msg[-1]
            raise RuntimeError(error_msg)

        final_relaxed_struct = LammpsData.from_file("final.relaxed", atom_style="charge").structure

        lines = [f"{final_relaxed_struct.num_sites}"]

        for idx, site in enumerate(final_relaxed_struct):
            if idx == final_idx:
                idx = final_relaxed_struct.num_sites
            elif idx == start_idx:
                idx = final_idx
            lines.append(f"{idx + 1}  {site.x:.3f}  {site.y:.3f}  {site.z:.3f}")

        with open("data.final_replica", "w") as f:
            f.write("\n".join(lines))

        input_file = "in.neb"

        with open(input_file, "w") as f:
            f.write(
                neb_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    start_replica="initial.relaxed",
                    final_replica="data.final_replica",
                )
            )

        return input_file

    def calculate(self):
        """Calculate the NEB barrier given Potential class."""
        with ScratchDir("."):
            input_file = self._setup()
            with subprocess.Popen(
                [
                    "mpirun",
                    "-n",
                    str(self.num_replicas),
                    self.LMP_EXE,
                    "-partition",
                    f"{self.num_replicas}x1",
                    "-in",
                    input_file,
                ],
                stdout=subprocess.PIPE,
            ) as p:
                stdout = p.communicate()[0]
                rc = p.returncode
            if rc != 0:
                error_msg = f"LAMMPS exited with return code {rc}"
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    logger.info(f"NudgedElasticBand error with message {msg}")
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)
            return self._parse()

    def _sanity_check(self, structure):
        """Check if the structure is valid for this calculation."""
        return True

    def _parse(self):
        """Parse results from dump files."""
        with open("log.lammps") as f:
            lines = f.readlines()[-1:]
        return float(lines[0].split()[6])


class DefectFormation(LMPStaticCalculator):
    """Defect formation energy calculator."""

    def __init__(self, ff_settings, specie, lattice, alat, **kwargs):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
            **kwargs: Passthrough to init.
        """
        self.ff_settings = ff_settings
        self.specie = specie
        self.lattice = lattice
        self.alat = alat
        super().__init__(**kwargs)

    @staticmethod
    def get_unit_cell(specie, lattice, alat):
        """
        Get the unit cell from specie, lattice type and lattice constant.

        Args:
            specie (str): Name of specie.
            lattice (str): The lattice type of structure. e.g. bcc or diamond.
            alat (float): The lattice constant of specific lattice and specie.
        """
        if lattice == "fcc":
            unit_cell = Structure.from_spacegroup(
                sg="Fm-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "bcc":
            unit_cell = Structure.from_spacegroup(
                sg="Im-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        elif lattice == "diamond":
            unit_cell = Structure.from_spacegroup(
                sg="Fd-3m", lattice=Lattice.cubic(alat), species=[specie], coords=[[0, 0, 0]]
            )
        else:
            raise ValueError("Lattice type is invalid.")

        return unit_cell

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "defect")

        with open(os.path.join(template_dir, "in.defect")) as f:
            defect_template = f.read()

        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=self.alat)
        lattice_calculator = LatticeConstant(ff_settings=self.ff_settings)
        a, _, _ = lattice_calculator.calculate([unit_cell])[0]
        unit_cell = self.get_unit_cell(specie=self.specie, lattice=self.lattice, alat=a)

        if self.lattice == "fcc":
            idx, scale_factor = 95, [3, 3, 3]
        elif self.lattice == "bcc":
            idx, scale_factor = 40, [3, 3, 3]
        elif self.lattice == "diamond":
            idx, scale_factor = 7, [2, 2, 2]
        else:
            raise ValueError("Lattice type is invalid.")

        super_cell = unit_cell * scale_factor
        efs_calculator = EnergyForceStress(ff_settings=self.ff_settings)
        energy_per_atom = efs_calculator.calculate([super_cell])[0][0] / len(super_cell)

        super_cell_ld = LammpsData.from_structure(
            super_cell, ff_elements=self.ff_settings.elements, atom_style="charge"
        )
        super_cell_ld.write_file("data.supercell")

        input_file = "in.defect"

        with open(input_file, "w") as f:
            f.write(
                defect_template.format(
                    ff_settings="\n".join(self.ff_settings.write_param()),
                    lattice=self.lattice,
                    alat=a,
                    specie=self.specie,
                    del_id=idx + 1,
                    relaxed_file="data.relaxed",
                )
            )

        return input_file, energy_per_atom, len(super_cell) - 1

    def calculate(self):
        """Calculate the vacancy formation given Potential class."""
        with ScratchDir("."):
            input_file, energy_per_atom, num_atoms = self._setup()
            with subprocess.Popen([self.LMP_EXE, "-in", input_file], stdout=subprocess.PIPE) as p:
                stdout = p.communicate()[0]
                rc = p.returncode
            if rc != 0:
                error_msg = f"LAMMPS exited with return code {rc}"
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)
            defect_energy, _, _ = self._parse()
        return defect_energy - energy_per_atom * num_atoms

    def _sanity_check(self, structure):
        return True

    def _parse(self):
        energy = float(np.loadtxt("energy.txt"))
        force = _read_dump("force.dump")
        stress = np.loadtxt("stress.txt")
        return energy, force, stress


class LMPRelaxationCalculator(LMPStaticCalculator):
    """Structural Relaxation Calculator."""

    def __init__(
        self,
        ff_settings,
        box_relax=True,
        box_relax_keywords="aniso 0.0 vmax 0.001",
        box_triclinic=False,
        min_style="cg",
        etol=1e-15,
        ftol=1e-15,
        maxiter=5000,
        maxeval=5000,
        **kwargs,
    ):
        """
        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            box_relax (bool): Whether to allow the box size and shape to vary
                during the minimization.
            box_relax_keywords (str): Keywords and values to define the conditions
                and constraints on the box size and shape.
            box_triclinic (bool): Whether to change box to triclinic so that
                LAMMPS can adjust x, y and z independently in relaxation.
            min_style (str): The minimization algorithm to use.
            etol (float): The stopping tolerance for energy during the minimization.
            ftol (float): The stopping tolerance for force during the minimization.
            maxiter (int): The max iterations of minimizer.
            maxeval (int): The max number of force/energy evaluations.
            **kwargs: Passthrough to init.
        """
        self.ff_settings = ff_settings
        self.box_relax = box_relax
        self.box_relax_keywords = box_relax_keywords
        self.box_triclinic = box_triclinic
        self.min_style = min_style
        self.etol = etol
        self.ftol = ftol
        self.maxiter = maxiter
        self.maxeval = maxeval
        super().__init__(**kwargs)

    def _setup(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "relax")

        with open(os.path.join(template_dir, "in.relax")) as f:
            input_template = f.read()

        input_file = "in.relax"

        ff_settings = self.ff_settings.write_param() if isinstance(self.ff_settings, Potential) else self.ff_settings

        box_relax_settings = ""
        if self.box_relax:
            box_relax_settings = f"fix   1 all box/relax {self.box_relax_keywords}"

        change_box = "" if not self.box_triclinic else "change_box all triclinic"

        inputs = input_template.format(
            ff_settings="\n".join(ff_settings),
            box_relax_settings=box_relax_settings,
            change_box=change_box,
            min_style=self.min_style,
            etol=self.etol,
            ftol=self.ftol,
            maxiter=self.maxiter,
            maxeval=self.maxeval,
        )

        with open(input_file, "w") as f:
            f.write(inputs)

        return input_file

    def _sanity_check(self, structure):
        """Check if the structure is valid for this calculation."""
        return True

    def _parse(self):
        """Parse results from dump files."""
        ld = LammpsData.from_file("data.relaxed", atom_style="charge")
        final_structure = ld.structure
        efs_calculator = EnergyForceStress(ff_settings=self.ff_settings)
        energy, forces, stresses = efs_calculator.calculate([final_structure])[0]
        return final_structure, energy, forces, stresses


class LatticeConstant(LMPRelaxationCalculator):
    """Lattice Constant Relaxation Calculator."""

    def calculate(self, structures):
        """
        Calculate the relaxed lattice parameters of a list of structures.

        Args:
            structures ([Structure]): Input structures in a list.

        Returns:
            List of relaxed lattice constants (a, b, c in Å) of the input structures.

        """
        results = super().calculate(structures)
        structures_relaxed = [r[0] for r in results]
        return [list(struct.lattice.abc) for struct in structures_relaxed]


class SurfaceEnergy(LMPRelaxationCalculator):
    """
    Surface energy Calculator.

    This calculator generate and calculate surface energies of slab structures based on inputs of
    bulk_structure and miller_indexes with the SlabGenerator in pymatgen:
    https://pymatgen.org/pymatgen.core.surface.html

    """

    def __init__(
        self,
        ff_settings,
        bulk_structure,
        miller_indexes,
        min_slab_size=15,
        min_vacuum_size=15,
        lll_reduce=False,
        center_slab=False,
        in_unit_planes=False,
        primitive=True,
        max_normal_search=None,
        reorient_lattice=True,
        box_relax=False,
        **kwargs,
    ):
        """
        Init.

        Args:
            ff_settings (list/Potential): Configure the force field settings for
                LAMMPS calculation, if given a Potential object, should apply
                Potential.write_param method to get the force field setting.
            bulk_structure (Structure): The bulk structure of target system. Slab structures
                will be generated based on inputs of bulk_structure and miller_indexes with
                the SlabGenerator in pymatgen.
            miller_indexes (list): A list of miller indexes of planes parallel to
                the surface of target slabs. Slab structures
                will be generated based on inputs of bulk_structure and miller_indexes with
                the SlabGenerator in pymatgen.
            min_slab_size (float): Minimum size in angstroms of layers containing atoms.
            min_vacuum_size (float): Minimize size in angstroms of layers containing vacuum.
            lll_reduce (bool): Whether the slabs will be orthogonalized.
            center_slab (bool): Whether the slabs will be centered between the vacuum layer.
            in_unit_planes (bool): Whether to set min_slab_size and min_vac_size in units
                of hkl planes (True) or Angstrom (False/default).
            primitive (bool): Whether to reduce any generated slabs to a primitive cell.
            max_normal_search (int): If set to a positive integer, the code will conduct a search
                for a normal lattice vector that is as perpendicular to the surface as possible by
                considering multiples linear combinations of lattice vectors up to max_normal_search.
            reorient_lattice (bool): Whether to reorient the lattice parameters such that
                the c direction is the third vector of the lattice matrix.
            box_relax (bool): Whether to allow the box size and shape of the slab structures to vary
                during the minimization. Normally, this should be turned off.
            **kwargs: Passthrough to init.
        """
        super().__init__(ff_settings=ff_settings, box_relax=True, **kwargs)
        bulk_structure_relaxed, bulk_energy, _, _ = super().calculate([bulk_structure])[0]
        self.bulk_energy_per_atom = bulk_energy / bulk_structure_relaxed.num_sites
        from pymatgen.core.surface import SlabGenerator

        slab_generators = [
            SlabGenerator(
                initial_structure=bulk_structure_relaxed,
                miller_index=miller_index,
                min_slab_size=min_slab_size,
                min_vacuum_size=min_vacuum_size,
                lll_reduce=lll_reduce,
                center_slab=center_slab,
                in_unit_planes=in_unit_planes,
                primitive=primitive,
                max_normal_search=max_normal_search,
                reorient_lattice=reorient_lattice,
            )
            for miller_index in miller_indexes
        ]
        slabs = [SG.get_slab() for SG in slab_generators]
        self.miller_indexes = miller_indexes
        self.slabs = slabs
        self.surface_areas = [s.surface_area for s in slabs]

        super().__init__(ff_settings=ff_settings, box_relax=box_relax, **kwargs)

    def calculate(self):
        """
        Calculate surface energies with the formula:
        E(Surface) = (E(Slab) - E(bulk)) / Area(surface). (J/m^2).

        Returns:
            List of miller_indexes with their respective relaxed slab structures and surface energies in J/m^2.

        """
        results = super().calculate(self.slabs)
        slabs_relaxed = [r[0] for r in results]
        slab_energies = [r[1] for r in results]
        surface_energies = [
            (slab_energy - slabs_relaxed[i].num_sites * self.bulk_energy_per_atom) / 2 / self.surface_areas[i]
            for i, slab_energy in enumerate(slab_energies)
        ]
        from pymatgen.core.units import Unit

        energy_unit_conversion_factor = Unit("eV").get_conversion_factor("J")
        length_unit_conversion_factor = Unit("ang").get_conversion_factor("m")
        surface_energies = [
            surface_energy * energy_unit_conversion_factor / length_unit_conversion_factor**2
            for surface_energy in surface_energies
        ]

        return list(zip(self.miller_indexes, slabs_relaxed, surface_energies, strict=False))


class LammpsPotential(Potential, abc.ABC):  # type: ignore
    """
    Lammps compatible potentials that call lammps executable for
    energy/force/stress calculations.
    """

    def predict_efs(self, structure):
        """
        Predict energy, forces and stresses of the structure.

        Args:
            structure (Structure): Pymatgen Structure object.

        Returns:
            energy, forces, stress
        """
        calculator = EnergyForceStress(self)
        energy, forces, stress = calculator.calculate(structures=[structure])[0]
        return energy, forces, stress



"""Feature Generator."""

from __future__ import annotations

import warnings
from functools import partial
from itertools import combinations_with_replacement
from typing import TYPE_CHECKING, Any

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Callable

    import pandas as pd


def _update_df(df, op, fn1, fn2=None):
    """Helper function to update the dataframe with new generated feature array."""
    fnames = df.columns
    if op.is_unary:
        new_fname = op.gen_name(fn1)
        if new_fname not in fnames:
            if op.rep in ["sqrt", "log10"] and np.any(df[fn1] < 0):
                warnings.warn(
                    f"Data {fn1} Contains negative number, sqrt will return complex number. "
                    "Consider using abssqrt or abslog10"
                )
            df[new_fname] = df[fn1].apply(op)
    elif op.is_binary:
        new_fname = op.gen_name(fn1, fn2)
        if new_fname not in fnames:
            df[new_fname] = op(df[fn1], df[fn2])


def generate_feature(feature_df: pd.DataFrame, operators: list) -> pd.DataFrame:
    """
    Generate new features by applying operators to columns in feature_df.

    Args:
        feature_df(pd.DataFrame): dataframe of original features
        operators(list): list of str of operators (check Operator.support_op_rep for reference)

    Returns: dataframe of augmented features

    """
    fdf = feature_df.copy()
    if not np.all([(o in Operator.support_op_rep) for o in operators]):
        raise ValueError("Contain unsupported operators, check Operator.supported_op_rep")
    ops = [Operator.from_str(o) for o in operators]
    sop = [op for op in ops if op.is_unary]
    bop = [op for op in ops if op.is_binary]
    for fn1, fn2 in combinations_with_replacement(fdf.columns, r=2):
        if fn1 == fn2:
            if not sop:
                continue
            for op in sop:
                _update_df(fdf, op, fn1)
        else:
            if not bop:
                continue
            for op in bop:
                _update_df(fdf, op, fn1, fn2)
                if not op.commutative:
                    _update_df(fdf, op, fn2, fn1)
    return fdf


class FeatureGenerator:
    """FeatureGenerator class for feature augmentation before selection."""

    def __init__(self, feature_df: pd.DataFrame, operators: list):
        """

        Args:
            feature_df(pd.DataFrame): original features
            operators(list): list of operators(str).
        """
        self.fdf = feature_df
        self.operators = operators

    def augment(self, n: int = 1) -> np.ndarray:
        """
        Augment features
        Args:
            n(int): number of rounds of iteration.

        Returns: augmented dataframe

        """
        df = self.fdf.copy()
        for _ in range(n):
            ndf = generate_feature(df, self.operators)
            df = ndf
        return df


class Operator:
    """
    Operator class. Wrap math operators with more attributes including check
    is_unary, is_binary, and is_commutative, and generate name string
    for the output.
    """

    support_op_rep = [
        "^2",
        "^3",
        "sqrt",
        "abssqrt",
        "cbrt",
        "exp",
        "abs",
        "log10",
        "abslog10",
        "+",
        "-",
        "*",
        "/",
        "|+|",
        "|-|",
        "sum_power_2",
        "sum_exp",
    ]

    def __init__(self, operation: Callable[..., Any], rep: str, unary: bool, commutative: bool):
        """
        Args:
            operation(Callable): operation function
            rep(str): representations of the operator
            unary(bool): whether it is a unary operator
            commutative(bool): whether it is a commutative operator.
        """
        self.opt = operation
        self.rep = rep
        self.unary = unary
        self.commutative = commutative

    def compute(self, i1: np.ndarray, i2: np.ndarray | None = None) -> np.ndarray:
        """
        Compute the results
        Args:
            i1(np.ndarray): first input array
            i2(np.ndarray): second input array (for binary operators).

        Returns: array of computed results
        """
        if self.is_binary:
            return self.opt(i1, i2)
        return self.opt(i1)

    def gen_name(self, f1: str, f2: str | None = None) -> str:
        """
        Generate string representation for output
        Args:
            f1(str): name of the first input array
            f2(str): name of the second input array.

        Returns: name of the output

        """
        assert self.rep in operation_dict
        f_format = str(operation_dict[self.rep]["f_format"])
        if self.is_binary:
            return f_format.format(f1=f1, f2=f2)
        return f_format.format(f1=f1)

    @classmethod
    def from_str(cls, op_name: str):
        """
        Operator from name of the operator
        Args:
            op_name(str): string representation of the operator,
            check Operator.support_op_rep for reference.

        Returns: Operator

        """
        kwgs = operation_dict[op_name]["kwgs"]
        return cls(**kwgs)

    @property
    def is_unary(self) -> bool:
        """Returns: True if the operator takes one argument else False."""
        if not self.unary:
            if self.rep in ["^2", "^3", "sqrt", "abssqrt", "cbrt", "exp", "abs", "log10", "abslog10"]:
                self.unary = True
            elif self.rep in ["+", "-", "*", "/", "|+|", "|-|", "sum_power_2", "sum_exp"]:
                self.unary = False
        return self.unary

    @property
    def is_binary(self) -> bool:
        """Returns: True if the operator takes two arguments else False."""
        return not self.is_unary

    @property
    def is_commutative(self) -> bool:
        """Returns: True if the operator is commutative else False."""
        if not self.commutative:
            self.commutative = not (self.is_unary or self.rep in ["-", "/"])
        return self.commutative

    def __call__(self, i1: np.ndarray, i2: np.ndarray | None = None) -> np.ndarray:
        """
        Compute the results
        Args:
            i1(np.ndarray): first input array
            i2(np.ndarray): second input array (for binary operators).

        Returns: array of computed results

        """
        return self.compute(i1, i2)

    def __str__(self) -> str:
        return self.rep


def _my_power(x: float, n: float) -> float:
    return pow(x, n)


def _my_abs_sqrt(x):
    return np.sqrt(abs(x))


def _my_exp(x):
    return np.exp(x)


def _my_abs_log10(x):
    return np.log10(abs(x))


def _my_exp_power_2(x):
    return np.exp(pow(x, 2))


def _my_exp_power_3(x):
    return np.exp(pow(x, 3))


def _my_sum(x, y):
    return x + y


def _my_abs_sum(x, y):
    return abs(x + y)


def _my_mul(x, y):
    return x * y


def _my_diff(x, y):
    return x - y


def _my_abs_diff(x, y):
    return abs(x - y)


def _my_div(x, y):
    return x / y


def _my_sum_power_2(x, y):
    return pow((x + y), 2)


def _my_sum_power_3(x, y):
    return pow((x + y), 3)


def _my_sum_exp(x, y):
    return np.exp(x + y)


# def _my_sum_exp_power_2(x, y):
#     return np.exp(pow(x + y, 2))
#
#
# def _my_sum_exp_power_3(x, y):
#     return np.exp(pow(x + y, 3))


operation_dict: dict[str, Any]
operation_dict = {
    "^2": {
        "kwgs": {"operation": partial(_my_power, n=2), "rep": "^2", "unary": True, "commutative": False},
        "f_format": "({f1})^2",
    },
    "^3": {
        "kwgs": {"operation": partial(_my_power, n=3), "rep": "^3", "unary": True, "commutative": False},
        "f_format": "({f1})^3",
    },
    "abs": {"kwgs": {"operation": abs, "rep": "abs", "unary": True, "commutative": False}, "f_format": "abs({f1})"},
    "sqrt": {
        "kwgs": {"operation": partial(_my_power, n=1 / 2), "rep": "sqrt", "unary": True, "commutative": False},
        "f_format": "sqrt({f1})",
    },
    "abssqrt": {
        "kwgs": {"operation": _my_abs_sqrt, "rep": "abssqrt", "unary": True, "commutative": False},
        "f_format": "sqrt(|{f1}|)",
    },
    "cbrt": {
        "kwgs": {"operation": partial(_my_power, n=1 / 3), "rep": "cbrt", "unary": True, "commutative": False},
        "f_format": "cbrt({f1})",
    },
    "exp": {"kwgs": {"operation": _my_exp, "rep": "exp", "unary": True, "commutative": False}, "f_format": "exp({f1})"},
    "log10": {
        "kwgs": {"operation": np.log10, "rep": "log10", "unary": True, "commutative": False},
        "f_format": "log10({f1})",
    },
    "abslog10": {
        "kwgs": {"operation": _my_abs_log10, "rep": "abslog10", "unary": True, "commutative": False},
        "f_format": "log10(|{f1}|)",
    },
    "+": {
        "kwgs": {"operation": _my_sum, "rep": "+", "unary": False, "commutative": True},
        "f_format": "(({f1}) + ({f2}))",
    },
    "|+|": {
        "kwgs": {"operation": _my_abs_sum, "rep": "|+|", "unary": False, "commutative": True},
        "f_format": "|({f1}) + ({f2})|",
    },
    "-": {
        "kwgs": {"operation": _my_diff, "rep": "-", "unary": False, "commutative": False},
        "f_format": "(({f1}) - ({f2}))",
    },
    "|-|": {
        "kwgs": {"operation": _my_abs_diff, "rep": "|-|", "unary": False, "commutative": True},
        "f_format": "|({f1}) - ({f2})|",
    },
    "*": {
        "kwgs": {"operation": _my_mul, "rep": "*", "unary": False, "commutative": True},
        "f_format": "(({f1}) * ({f2}))",
    },
    "/": {
        "kwgs": {"operation": _my_div, "rep": "/", "unary": False, "commutative": False},
        "f_format": "(({f1}) / ({f2}))",
    },
    "sum_power_2": {
        "kwgs": {"operation": _my_sum_power_2, "rep": "sum_power_2", "unary": False, "commutative": True},
        "f_format": "(({f1}) + ({f2}))^2",
    },
    "sum_exp": {
        "kwgs": {"operation": _my_sum_exp, "rep": "sum_exp", "unary": False, "commutative": True},
        "f_format": "exp(({f1}) + ({f2}))",
    },
}
